# -*- coding: utf-8 -*-
"""
###################################################################
#  Automator for Zorro and 2dx
#
#      License: LGPL
#      Author:  Robert A. McLeod
#      Website: https://github.com/C-CINA/zorro-dev
#
#  See LICENSE.txt for details about copyright and
#  rights to use.
####################################################################

This is a new PySide/Qt4 Gui for Zorro and 2dx.  It's capable of running multiple subprocesses at the same 
time, without resorting to the buggy Python multiprocessing module.  

####################### BUILDING PYSIDE FROM GIT
http://pyside.readthedocs.org/en/latest/building/linux.html
#######################
"""

import os, os.path, sys, glob

try:
    from PySide import QtGui, QtCore
    os.environ.setdefault('QT_API','pyside')
except:
    # Import PyQt4 as backup?  I suspect this still causes license issues
    ImportError( "Automator.py: PySide not found, I am going to crash now. Bye." )

import zorro
import zorroSkulkManager

# from copy import deepcopy
from Ui_Automator import Ui_Automator_ui
from Ui_DialogFileLoc import Ui_DialogFileLocations

# import time
import functools

import Queue
import ConfigParser
import numpy as np
import subprocess as sp

progname = os.path.basename(sys.argv[0])
progversion = "0.3.0"
extensionlist = ['*.dm4', '*.dm3', '*.mrc', '*.mrcs', '*.hdf5', '*.h5', 
                 '*.dm4.bz2', '*.dm3.bz2', '*.mrc.bz2', '*.mrcs.bz2', '*.hdf5.bz2', '*.h5.bz2', 
                 '*.dm4.gz', '*.dm3.gz', '*.mrc.gz', '*.mrcs.gz', '*.hdf5.gz', '*.h5.gz']

#class Automator(Ui_Automator_ui, QtGui.QApplication):
class Automator(Ui_Automator_ui, QtGui.QApplication):
    
    def __init__(self):
        # Hack to get icons to show up...
        origdir = os.path.realpath('.')
        os.chdir( os.path.dirname(os.path.realpath(__file__)) )

        QtGui.QApplication.__init__(self, sys.argv)
        #self.app = QtGui.QApplication(sys.argv)
        self.MainWindow = QtGui.QMainWindow()
        self.ImageDialog = QtGui.QDialog()
        self.FileLocDialog = QtGui.QDialog()

        self.setupUi(self.MainWindow)
        
        # Zorro objects and the default one built from the GUI
        self.zorroDefault = zorro.ImageRegistrator()
        
        self.skulkMessanger = Queue.Queue()
        self.skulk = zorroSkulkManager.skulkManager( self.skulkMessanger )
        # Connect skulk.automatorSignal to myself
        self.skulk.automatorSignal.connect( self.updateFromSkulk )
        self.skulkThread = None

        # Automator parameters
        # Store paths in a dict, raw, av, and align can be relative to out
        # Ramdisk may or may not be used.
        self.paths = {}
        self.paths['input_dir'] = ""
        self.paths['output_dir'] = ""
        self.paths['raw_subdir'] = ""
        self.paths['sum_subdir'] = ""
        self.paths['align_subdir'] = ""
        self.paths['figure_subdir'] = ""
        self.paths['output_format'] = "MRC"

        # self.paths['Use RAMdisk'] = False
        # self.paths['Size RAMdisk'] = 0.0
        
        # This is wrong somehow...
        # Now getting the dimensions from the parent correctly. 
        # self.figureV_2 = Twodx_MplCanvas.TwodxPlotMplCanvas(self.frameBottomV_2, width=3.0, height=3.0, dpi=50)
        
        self.cfgCommon = {}
        self.cfgCluster = {}
        self.cfgGauto = {}
        self.cfgGplot = {}
        
        # All the Gauto's are line-edits so that we can have None or "" as the values
        # Plus I don't have to worry about types.
        self.leGautoBoxsize.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'boxsize', self.leGautoBoxsize.text ) )
        self.leGautoDiameter.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'diameter', self.leGautoDiameter.text ) )    
        self.leGautoMin_Dist.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'min_dist', self.leGautoMin_Dist.text ) )        
        # Template
        self.leGautoTemplates.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'T', self.leGautoTemplates.text ) )
        # Template pixelsize?    
        self.leGautoAng_Step.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'ang_step', self.leGautoAng_Step.text ) )      
        self.leGautoSpeed.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'speed', self.leGautoSpeed.text ) ) 
        self.leGautoCCCutoff.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'cc_cutoff', self.leGautoCCCutoff.text ) ) 
        self.leGautoLsigma_D.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'lsigma_D', self.leGautoLsigma_D.text ) ) 
        self.leGautoLsigma_Cutoff.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'lsigma_cutoff', self.leGautoLsigma_Cutoff.text ) ) 
        self.leGautoLave_D.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'lave_D', self.leGautoLave_D.text ) ) 
        self.leGautoLave_Max.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'lave_max', self.leGautoLave_Max.text ) ) 
        self.leGautoLave_Min.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'lave_min', self.leGautoLave_Min.text ) ) 
        self.leGautoLP.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'lp', self.leGautoLP.text ) ) 
        self.leGautoHP.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'hp', self.leGautoHP.text ) ) 
        self.leGautoLPPre.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'pre_lp', self.leGautoLPPre.text ) ) 
        self.leGautoHPPre.editingFinished.connect( functools.partial( 
            self.updateDict, self.cfgGauto, 'pre_hp', self.leGautoHPPre.text ) ) 
            
        # Flags go into cfgGPlot, simply so I can use a generator on cfgGauto and handle these manually
        self.cbGautoDoprefilter.stateChanged.connect( functools.partial( 
            self.updateDict, self.cfgGplot, "do_pre_filter", self.cbGautoDoprefilter.isChecked ) )
        self.cbGautoPlotCCMax.stateChanged.connect( functools.partial( 
            self.updateDict, self.cfgGplot, "write_ccmax_mic", self.cbGautoPlotCCMax.isChecked ) )
        self.cbGautoPlotPref.stateChanged.connect( functools.partial( 
            self.updateDict, self.cfgGplot, "write_pref_mic", self.cbGautoPlotPref.isChecked ) )
        self.cbGautoPlotBG.stateChanged.connect( functools.partial( 
            self.updateDict, self.cfgGplot, "write_bg_mic", self.cbGautoPlotBG.isChecked ) )
        self.cbGautoPlotBGFree.stateChanged.connect( functools.partial( 
            self.updateDict, self.cfgGplot, "write_bgfree_mic", self.cbGautoPlotBGFree.isChecked ) )
        self.cbGautoPlotLsigma.stateChanged.connect( functools.partial( 
            self.updateDict, self.cfgGplot, "write_lsigma_mic", self.cbGautoPlotLsigma.isChecked ) )    
        self.cbGautoPlotMask.stateChanged.connect( functools.partial( 
            self.updateDict, self.cfgGplot, "write_mic_mask", self.cbGautoPlotMask.isChecked ) )     

        self.updateFileList()

        self.statusbar.showMessage( "Welcome to Automator for Zorro and 2dx, version " + progversion )
        
        self.ui_FileLocDialog = Ui_DialogFileLocations()
        self.ui_FileLocDialog.setupUi( self.FileLocDialog )
        
        
        # Apply all the connections from the QT gui objects to their associated functions in this class, using functools as desired
        # Menu items
        self.actionQuit.triggered.connect( self.quitApp )
        self.MainWindow.closeEvent = self.quitApp
        self.actionLoad_config.triggered.connect( functools.partial(self.loadConfig, None) )
        self.actionSave_config.triggered.connect( functools.partial(self.saveConfig, None) )
        self.actionSet_paths.triggered.connect( self.FileLocDialog.show )
        
        # Paths
        self.ui_FileLocDialog.tbOpenInputPath.clicked.connect( functools.partial( 
            self.openPathDialog, 'input_dir', True ) )
        self.ui_FileLocDialog.tbOpenOutputPath.clicked.connect( functools.partial( 
            self.openPathDialog, 'output_dir', True ) )
        self.ui_FileLocDialog.tbOpenRawPath.clicked.connect( functools.partial( 
            self.openPathDialog, 'raw_subdir', True ) )
        self.ui_FileLocDialog.tbOpenSumPath.clicked.connect( functools.partial( 
            self.openPathDialog, 'sum_subdir', True ) )
        self.ui_FileLocDialog.tbOpenAlignPath.clicked.connect( functools.partial( 
            self.openPathDialog, 'align_subdir', True ) )
        self.ui_FileLocDialog.tbOpenFiguresPath.clicked.connect( functools.partial( 
            self.openPathDialog, 'figure_subdir', True ) )
        self.ui_FileLocDialog.comboOutputFormat.currentIndexChanged.connect( functools.partial( 
            self.updateDict, self.paths, 'output_format', self.ui_FileLocDialog.comboOutputFormat.currentText ) )
            
        self.ui_FileLocDialog.leInputPath.editingFinished.connect( functools.partial( 
            self.updateDict, self.paths, 'input_dir', self.ui_FileLocDialog.leInputPath.text ) )
        self.ui_FileLocDialog.leOutputPath.editingFinished.connect( functools.partial( 
            self.updateDict, self.paths, 'output_dir', self.ui_FileLocDialog.leOutputPath.text ) )
        self.ui_FileLocDialog.leRawPath.editingFinished.connect( functools.partial( 
            self.updateDict, self.paths, 'raw_subdir', self.ui_FileLocDialog.leRawPath.text ) )
        self.ui_FileLocDialog.leSumPath.editingFinished.connect( functools.partial( 
            self.updateDict, self.paths, 'sum_subdir', self.ui_FileLocDialog.leSumPath.text ) ) 
        self.ui_FileLocDialog.leAlignPath.editingFinished.connect( functools.partial( 
            self.updateDict, self.paths, 'align_subdir', self.ui_FileLocDialog.leAlignPath.text ) )
        self.ui_FileLocDialog.leFiguresPath.editingFinished.connect( functools.partial( 
            self.updateDict, self.paths, 'figure_subdir', self.ui_FileLocDialog.leFiguresPath.text ) )    
        
        # Common configuration
        self.comboRegistrator.currentIndexChanged.connect( functools.partial( 
            self.updateDict, self.cfgCommon, "registrator", self.comboRegistrator.currentText ) )
        self.comboTriMode.currentIndexChanged.connect( functools.partial( 
            self.updateZorroDefault, "triMode", self.comboTriMode.currentText ) )
        self.sbPeaksigThres.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "peaksigThres", self.sbPeaksigThres.value ) )    
#        self.sbStartFrame.valueChanged.connect( functools.partial( 
#            self.updateZorroDefault, "startFrame", self.sbStartFrame.value ) )    
#        self.sbEndFrame.valueChanged.connect( functools.partial( 
#            self.updateZorroDefault, "endFrame", self.sbEndFrame.value ) )    
        self.sbDiagWidth.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "diagWidth", self.sbDiagWidth.value ) )        
        self.sbAutomax.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "autoMax", self.sbAutomax.value ) )    
        self.cbSuppressOrigin.stateChanged.connect( functools.partial( 
            self.updateZorroDefault, "suppressOrigin", self.cbSuppressOrigin.isChecked ) )       
        self.comboCompressionExt.currentIndexChanged.connect( functools.partial( 
            self.updateZorroDefault, "compress_ext", self.comboCompressionExt.currentText ) )    
        self.cbSavePNG.stateChanged.connect( functools.partial( 
            self.updateZorroDefault, "savePNG", self.cbSavePNG.isChecked ) )
        self.cbDoseFilter.stateChanged.connect( functools.partial( 
            self.updateZorroDefault, "doDoseFilter", self.cbDoseFilter.isChecked ) )
        self.cbDoCompression.stateChanged.connect( functools.partial( 
            self.updateZorroDefault, "doCompression", self.cbDoCompression.isChecked ) )  
        self.cbDoCTF.stateChanged.connect( functools.partial( 
            self.updateZorroDefault, "doCTF", self.cbDoCTF.isChecked ) )      
            
        # Advanced configuration
        self.sbShapePadX.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "shapePadded", (self.sbShapePadY.value,self.sbShapePadX.value) ) )
        self.sbShapePadY.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "shapePadded", (self.sbShapePadY.value,self.sbShapePadX.value) ) )   
        self.sbFouCropX.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "fouCrop", (self.sbFouCropY.value,self.sbFouCropX.value) ) )
        self.sbFouCropY.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "fouCrop", (self.sbFouCropY.value,self.sbFouCropX.value) ) )  
        self.sbPixelsize.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "pixelsize", self.sbPixelsize.value ) )       
        self.sbVoltage.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "voltage", self.sbVoltage.value ) )  
        self.sbC3.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "C3", self.sbC3.value ) )
        self.sbMaxShift.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "maxShift", self.sbMaxShift.value ) )    
        self.comboOriginMode.currentIndexChanged.connect( functools.partial( 
            self.updateZorroDefault, "originMode", self.comboOriginMode.currentText ) )    
        self.cbPreshift.stateChanged.connect( functools.partial( 
            self.updateZorroDefault, "preShift", self.cbPreshift.isChecked ) )  
        self.cbSaveC.stateChanged.connect( functools.partial( 
            self.updateZorroDefault, "saveC", self.cbSaveC.isChecked ) )      
        self.comboBmode.currentIndexChanged.connect( functools.partial( 
            self.updateZorroDefault, "Bmode", self.comboBmode.currentText ) ) 
        self.sbBrad.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "Brad", self.sbBrad.value ) )    
        self.comboWeightMode.currentIndexChanged.connect( functools.partial( 
            self.updateZorroDefault, "weightMode", self.comboWeightMode.currentText ) )     
        self.comboPeakLocMethod.currentIndexChanged.connect( functools.partial( 
            self.updateZorroDefault, "peakLocMode", self.comboPeakLocMethod.currentText ) )  
        self.sbSubpixReg.valueChanged.connect( functools.partial( 
            self.updateZorroDefault, "subPixReg", self.sbSubpixReg.value ) )  
        self.comboShiftMethod.currentIndexChanged.connect( functools.partial( 
            self.updateZorroDefault, "shiftMethod", self.comboShiftMethod.currentText ) )  
            
        # Cluster configuration
        self.comboClusterType.currentIndexChanged.connect( functools.partial( 
            self.updateDict, self.cfgCluster, "cluster_type", self.comboClusterType.currentText ) )

        self.sbNThreads.valueChanged.connect( functools.partial( 
            self.updateDict, self.cfgCluster, 'n_threads', self.sbNThreads.value ) )

        self.sbNProcesses.valueChanged.connect( functools.partial( 
            self.updateDict, self.cfgCluster, "n_processes", self.sbNProcesses.value ) )
        self.leCachePath.textEdited.connect( functools.partial( 
            self.updateZorroDefault, 'cachePath', self.leCachePath.text ))
        self.comboFFTWEffort.currentIndexChanged.connect( functools.partial( 
            self.updateZorroDefault,  "fftw_effort", self.comboFFTWEffort.currentText ) )    

        self.listFiles.itemActivated.connect( self.checkSelectedFile  )
        
        self.tbDeleteFile.clicked.connect( self.deleteSelectedFile )
        self.tbRun.clicked.connect( self.runSkulk )
        self.tbKillAll.clicked.connect( self.killSkulk )
        self.tbKillAll.setEnabled(False)
        self.tbReprocess.clicked.connect( self.reprocessSelectedFile )
        self.tbParticlePick.clicked.connect( self.particlePick )

        # Plots setup
        # What I can do is make a list of viewWidgets so I can iterate through them
        # This gives more flexibility for different arrangements in the future
        self.viewWidgetList = []
        self.viewWidgetList.append( self.viewWidget1 )
        self.viewWidgetList.append( self.viewWidget2 )
        self.viewWidgetList.append( self.viewWidget3 )
        self.viewWidgetList.append( self.viewWidget4 )
        self.viewWidgetList.append( self.viewWidget5 )
        self.viewWidgetList.append( self.viewWidget6 )
        for index, viewWidg in enumerate( self.viewWidgetList ):
            viewWidg.viewNumber = index
            viewWidg.autoParent = self
        
        # Check for installation of lbzip2, pigz, and 7z
        if not bool( zorro.util.which('lbzip2') ) and not bool( zorro.util.which('7z') ) and not bool( zorro.util.which('pigz') ):
            warnBox = QtGui.QMessageBox()
            warnBox.setText( "Disabling compression: None of lbzip2, pigz, or 7z found" )
            # TODO: limit compress_ext if only one is found?
            self.comboCompressionExt.setEnabled(False)
            self.cbDoCompression.setChecked(False)
            self.cbDoCompression.setEnabled(False)
            warnBox.exec_()
        
        # Check for installation of CTFFIND
        if not bool( zorro.util.which('ctffind') ):
            warnBox = QtGui.QMessageBox()
            warnBox.setText( "Disabling CTF measurement: CTFFIND4 not found" )
            self.cbDoCTF.setChecked(False)
            self.cbDoCTF.setEnabled(False)
            warnBox.exec_()
            
        if not bool( zorro.util.which('Gautomatch') ):
            warnBox = QtGui.QMessageBox()
            warnBox.setText( "Disabling Gautomatch measurement: Gautomatch not found" )
            self.pageGautoConfig.setEnabled(False)
            self.tbParticlePick.setEnabled(False)
            warnBox.exec_()

        # self.MainWindow.showFullScreen()
        self.MainWindow.showMaximized()
        # self.FileLocDialog.show()
        # self.quitApp()
        
        # Try to load an ini file in the startup directory if it's present.
        try: 
            iniList = glob.glob( os.path.join( origdir, '*.ini' ) )
            # Can't open a loadConfig dialog until the app has started, so only one .ini file can be in the directory.
            if len( iniList ) == 1: # One .ini file
                self.loadConfig( iniList[0] )
            else:
                self.loadConfig( 'default.ini' )
        except:
            self.loadConfig( 'default.ini' )
            
        os.chdir(origdir)
        # self.app.exec_()
        
        self.skulk.inspectOutputDirectory() # Let's see if we can run this once...
        self.exec_()
        
        
        
    def initSkulk( self ):
        # TODO: this maybe should be inside the zorroSkulkManager
        self.skulk.hostInfo = {}
        
        self.cfgCluster['n_threads']
        if self.cfgCluster['cluster_type'] == 'local':
            for J in xrange( self.cfgCluster['n_processes'] ):
                self.skulk.hostInfo['localhost%d'%J]= zorroSkulkManager.skulkHost( 'localhost%d'%J, 'local', 
                    self.skulk.hostMessageQueue, n_threads=self.cfgCluster['n_threads'] )
        
        
    def runSkulk( self ):
        # Clean-up the file list
        for J in np.arange( self.listFiles.count(), -1, -1 ):
            self.listFiles.takeItem( J )
            
        # Check pathing, ask user to set if any fields are missing
        if not self.checkValidPaths():
            return
            
        # Init
        self.skulk.inspectOutputDirectory()
        self.skulk._skulkManager__STOP_POLLING = False
        self.initSkulk()
        
        # Make a new thread
        self.skulk.start()
        
        # Disable GUI elements related to clusters
        self.tbKillAll.setEnabled(True)
        self.tbRun.setEnabled(False)
        self.sbNProcesses.setEnabled(False)
        self.sbNThreads.setEnabled(False)
        self.comboClusterType.setEnabled(False)
        
    def killSkulk( self ):
        print( "Killing all the foxes in the skulk" )
        self.skulk.kill()
        #if self.skulkThread != None:
        # self.skulkThread.join()
        #self.skulkThread.exit()
            
        # Re-enable locked UI elements
        self.tbKillAll.setEnabled(False)
        self.tbRun.setEnabled(True)
        self.sbNProcesses.setEnabled(True)
        self.sbNThreads.setEnabled(True)
        self.comboClusterType.setEnabled(True)
        

    @QtCore.Slot()    
    def updateFromSkulk( self, key, status ):
        """
        This is a message from the skulk manager that it's had a file change
        
        Valid colors:
        u'aliceblue', u'antiquewhite', u'aqua', u'aquamarine', u'azure', u'beige', u'bisque', u'black', 
        u'blanchedalmond', u'blue', u'blueviolet', u'brown', u'burlywood', u'cadetblue', u'chartreuse', 
        u'chocolate', u'coral', u'cornflowerblue', u'cornsilk', u'crimson', u'cyan', u'darkblue', u'darkcyan', 
        u'darkgoldenrod', u'darkgray', u'darkgreen', u'darkgrey', u'darkkhaki', u'darkmagenta', 
        u'darkolivegreen', u'darkorange', u'darkorchid', u'darkred', u'darksalmon', u'darkseagreen', 
        u'darkslateblue', u'darkslategray', u'darkslategrey', u'darkturquoise', u'darkviolet', u'deeppink', 
        u'deepskyblue', u'dimgray', u'dimgrey', u'dodgerblue', u'firebrick', u'floralwhite', u'forestgreen', 
        u'fuchsia', u'gainsboro', u'ghostwhite', u'gold', u'goldenrod', u'gray', u'green', u'greenyellow', 
        u'grey', u'honeydew', u'hotpink', u'indianred', u'indigo', u'ivory', u'khaki', u'lavender', 
        u'lavenderblush', u'lawngreen', u'lemonchiffon', u'lightblue', u'lightcoral', u'lightcyan', 
        u'lightgoldenrodyellow', u'lightgray', u'lightgreen', u'lightgrey', u'lightpink', u'lightsalmon', 
        u'lightseagreen', u'lightskyblue', u'lightslategray', u'lightslategrey', u'lightsteelblue', 
        u'lightyellow', u'lime', u'limegreen', u'linen', u'magenta', u'maroon', u'mediumaquamarine', 
        u'mediumblue', u'mediumorchid', u'mediumpurple', u'mediumseagreen', u'mediumslateblue', 
        u'mediumspringgreen', u'mediumturquoise', u'mediumvioletred', u'midnightblue', u'mintcream', 
        u'mistyrose', u'moccasin', u'navajowhite', u'navy', u'oldlace', u'olive', u'olivedrab', u'orange', 
        u'orangered', u'orchid', u'palegoldenrod', u'palegreen', u'paleturquoise', u'palevioletred', 
        u'papayawhip', u'peachpuff', u'peru', u'pink', u'plum', u'powderblue', u'purple', u'red', u'rosybrown', 
        u'royalblue', u'saddlebrown', u'salmon', u'sandybrown', u'seagreen', u'seashell', u'sienna', 
        u'silver', u'skyblue', u'slateblue', u'slategray', u'slategrey', u'snow', u'springgreen', u'steelblue', 
        u'tan', u'teal', u'thistle', u'tomato', u'transparent', u'turquoise', u'violet', u'wheat', u'white', 
        u'whitesmoke', u'yellow', u'yellowgreen']
        """
        print( "Updating from skulk " + key + " : " + status )
        
        # Check to see if the item exists already
        listItem = self.listFiles.findItems( key, QtCore.Qt.MatchFixedString )

        if len(listItem) == 0:
            listItem = QtGui.QListWidgetItem( key )
            self.listFiles.addItem( listItem )
        else:
            listItem = listItem[0]
            
        # Update color
        if status == 'fini': 
            self.statusbar.showMessage( "Finished processing: " + key )
            listItem.setForeground( QtGui.QBrush( QtGui.QColor( u"green" ) ) )
        elif status == 'new' or status=='changing':
            listItem.setForeground( QtGui.QBrush( QtGui.QColor( u"orange" ) ) )
        elif status == 'stable':
            listItem.setForeground( QtGui.QBrush( QtGui.QColor( u"firebrick" ) ) )
        elif status == 'proc':
            listItem.setForeground( QtGui.QBrush( QtGui.QColor( u"steelblue"  ) ) )
        elif status == 'remove':
            self.listFiles.takeItem( self.listFiles.row( listItem) )
        pass

        sizeFini = len( self.skulk.skulkFini )
        sizeTotal = len( self.skulk )
        self.labelMovie.setText( "Stack browser: %d"%sizeFini + " done / %d"%(sizeTotal)+ " tot" )

        
    def checkSelectedFile( self, item ):
        # Get the ZorroObj from the stack browser
        key = item.text()
        if self.skulk.skulkFini.has_key( key ):
            print( "TODO: found zorroObject ("+str(key)+") in the skulkManager" )
            print( self.skulk.skulkFini[key] )
            self.updateAllViews( zorroObj = self.skulk.skulkFini[key] )
        
    def deleteSelectedFile( self ):
        itemList = self.listFiles.selectedItems()
        confirmBox = QtGui.QMessageBox()
        filenameString = ""
        for item in itemList:
            filenameString += item.text() + "\n"
        confirmBox.setText( "Are you sure you want to delete all files related to: %s" % filenameString )
        confirmBox.addButton( QtGui.QMessageBox.Cancel )
        deleteButton = confirmBox.addButton( "Delete", QtGui.QMessageBox.ActionRole )
        confirmBox.setDefaultButton( QtGui.QMessageBox.Cancel )
        confirmBox.exec_()
        if confirmBox.clickedButton() == deleteButton:
            for item in itemList:
                #item = self.listFiles.currentItem()
                if item is None: continue

                # Delete everything
                # TODO: this may not work properly with the relative pathing.=
                print( "Delete: " + item.text() )
                self.skulk.remove( item.text() )
                self.listFiles.takeItem( self.listFiles.row(item) )
        
    def reprocessSelectedFile( self ):
        # item = self.listFiles.currentItem()
        itemList = self.listFiles.selectedItems()
        for item in itemList:
            if item is None: continue
            
            zorroObj = self.skulk[item.text()]
            if zorroObj != None:
                self.listFiles.takeItem( self.listFiles.row(item) )   
                print( "REPROCESS: %s" % item.text() )
                self.skulk.reprocess( item.text() )

    def particlePick( self ):
        import Gautoauto
        
        itemList = self.listFiles.selectedItems()
        
        sumList = []
        pngFronts = []
        for item in itemList:
            if item is None: continue
            
            # Get the zorro obj and feed the imageSum or filtSum to execGautoMatch
            # Maybe it would be safer to load the zorroObj explicitely?
            zorroObj = self.skulk[item.text()]
            
            if 'filt' in zorroObj.files:
                baseDir = os.path.split( zorroObj.files['config'] )[0]
                sumFilename = os.path.join( baseDir, zorroObj.files['filt'])
                sumList.append( sumFilename )     
                
            elif 'sum' in zorroObj.files:
                baseDir = os.path.split( zorroObj.files['config'] )[0]
                sumList.append( os.path.join( baseDir, zorroObj.files['sum']) )
                
            else:
                print( "Could not find image sum for %s to particle pick" % item.text() )
                try:
                    print( "zorroObj.files: " + str(zorroObj.files) )
                except:
                    pass
                return
                
            stackFront = os.path.splitext( os.path.split( zorroObj.files['stack'])[1] )[0]    
            pngDir = os.path.join( baseDir, zorroObj.files['figurePath'] )
            pngFileFront = os.path.join( pngDir, stackFront  )
            pngFronts.append( pngFileFront )
            # Add boxMask to the zorroObj files by default?
            zorroObj.files['figBoxMask'] = pngFileFront + "_boxMask.png"
            # TODO: add actual box file to files list
            # TODO: This also implies saving the config file
            # zorroObj.saveConfig(  zorroObj.files['config'] )
            

        # Submit job, this should be a new thread
        print( "===== TODO: launch Gautomatch in seperate thread as it blocks =====" )
        self.cfgGplot['edge'] = 64
        self.cfgGauto['apixM'] = 10.0*zorroObj.pixelsize # This assumes all micrographs in the same directory have the same PS
        self.cfgGplot['colorMap'] = 'viridis'
        self.cfgGplot['boxAlpha'] = 0.5
        self.cfgGplot['shapeOriginal'] = zorroObj.shapeOriginal
        self.cfgGplot['binning'] = 4
        Gautoauto.batchProcess( sumList, pngFronts, self.cfgGauto, self.cfgGplot, n_processes=self.sbGautoNProcesses.value() )
            

    
    def checkValidPaths( self ):
        # See if the directories exist.  If not, try to make them.  Zorro does this as well but it's 
        # better to have user feedback at the start.
        # self.paths['input_dir'] = ""; self.paths['output_dir'] = ""l self.paths['raw_subdir'] = ""
        # self.paths['sum_subdir'] = ""; self.paths['align_subdir'] = ""; self.paths['figure_subdir'] = ""
        errorText = ""
        errorState = False
        
        try:
            if not bool(self.paths['input_dir']):
                errorState = True; errorText += "Error: Input directory field is empty.\n"
                raise ValueError
            if not os.path.isdir( self.paths['input_dir'] ):
                os.mkdir( self.paths['input_dir'] )
            if not os.access( self.paths['input_dir'], os.R_OK ):
                errorState = True; errorText += "Error: Input directory has no read permissions.\n"
            if not os.access( self.paths['input_dir'], os.W_OK ):
                errorText += "Warning: Input directory has no write permissions, cannot delete stacks from SSD.\n"
        except OSError:
            errorState = True; errorText += "Error: Input directory does not exist and could not be made.\n"
        except ValueError: 
            pass
                
        try:
            if not bool(self.paths['output_dir']):
                errorState = True; errorText += "Error: Output directory field is empty.\n"
                raise ValueError
            if not os.path.isdir( self.paths['output_dir'] ):
                os.mkdir( self.paths['output_dir'] )
            if not os.access( self.paths['output_dir'], os.R_OK ):
                errorState = True; errorText += "Error: Output directory has no read permissions.\n"
            if not os.access( self.paths['output_dir'], os.W_OK ):
                errorState = True; errorText += "Warning: Output directory has no write permissions.\n"
        except OSError:
            errorState = True; errorText += "Error: Output directory does not exist and could not be made.\n"
        except ValueError: 
            pass
        
        try:
            if not bool(self.paths['raw_subdir']):
                self.paths['raw_subdir'] = os.path.join( self.paths['output_dir'], 'raw' )
                self.ui_FileLocDialog.leRawPath.setText( self.paths['raw_subdir'] ) # Doesn't fire event
                errorText += "Warning: Raw directory set to default <out>/raw.\n"
            if not os.path.isdir( self.paths['raw_subdir'] ):
                os.mkdir( self.paths['raw_subdir'] )
            if not os.access( self.paths['raw_subdir'], os.R_OK ):
                errorState = True; errorText += "Error: Raw directory has no read permissions.\n"
            if not os.access( self.paths['raw_subdir'], os.W_OK ):
                errorState = True; errorText += "Warning: Raw directory has no write permissions.\n"
        except OSError:
            errorState = True; errorText += "Error: Raw directory does not exist and could not be made.\n"

        try:
            if not bool(self.paths['sum_subdir']):
                self.paths['sum_subdir'] = os.path.join( self.paths['output_dir'], 'sum' )
                self.ui_FileLocDialog.leSumPath.setText( self.paths['sum_subdir'] ) # Doesn't fire event
                errorText += "Warning: sum directory set to default <out>/sum.\n"
            if not os.path.isdir( self.paths['sum_subdir'] ):
                os.mkdir( self.paths['sum_subdir'] )
            if not os.access( self.paths['sum_subdir'], os.R_OK ):
                errorState = True; errorText += "Error: sum directory has no read permissions.\n"
            if not os.access( self.paths['sum_subdir'], os.W_OK ):
                errorState = True; errorText += "Warning: sum directory has no write permissions.\n"
        except OSError:
            errorState = True; errorText += "Error: sum directory does not exist and could not be made.\n"

        try:
            if not bool(self.paths['align_subdir']):
                self.paths['align_subdir'] = os.path.join( self.paths['output_dir'], 'align' )
                self.ui_FileLocDialog.leAlignPath.setText( self.paths['align_subdir'] ) # Doesn't fire event
                errorText += "Warning: align directory set to default <out>/align.\n"
            if not os.path.isdir( self.paths['align_subdir'] ):
                os.mkdir( self.paths['align_subdir'] )
            if not os.access( self.paths['align_subdir'], os.R_OK ):
                errorState = True; errorText += "Error: align directory has no read permissions.\n"
            if not os.access( self.paths['align_subdir'], os.W_OK ):
                errorState = True; errorText += "Warning: align directory has no write permissions.\n"
        except OSError:
            errorState = True; errorText += "Error: align directory does not exist and could not be made.\n"
        
        try:
            if not bool(self.paths['figure_subdir']):
                self.paths['figure_subdir'] = os.path.join( self.paths['output_dir'], 'figure' )
                self.ui_FileLocDialog.leFiguresPath.setText( self.paths['figure_subdir'] ) # Doesn't fire event
                errorText += "Warning: figure directory set to default <out>/figure.\n"
            if not os.path.isdir( self.paths['figure_subdir'] ):
                os.mkdir( self.paths['figure_subdir'] )
            if not os.access( self.paths['figure_subdir'], os.R_OK ):
                errorState = True; errorText += "Error: figure directory has no read permissions.\n"
            if not os.access( self.paths['figure_subdir'], os.W_OK ):
                errorState = True; errorText += "Warning: figure directory has no write permissions.\n"
        except OSError:
            errorState = True; errorText += "Error: figure directory does not exist and could not be made.\n"
        
        # Check for path uniqueness
        if os.path.realpath(self.paths['input_dir']) == os.path.realpath(self.paths['output_dir']):
            errorState = True; errorText += "Error: Input and output directory may not be the same.\n"
        if os.path.realpath(self.paths['input_dir']) == os.path.realpath(self.paths['raw_subdir']):
            errorState = True; errorText += "Error: Input and raw directory may not be the same.\n"
        if os.path.realpath(self.paths['input_dir']) == os.path.realpath(self.paths['sum_subdir']):
            errorState = True; errorText += "Error: Input and sum directory may not be the same.\n"
        if os.path.realpath(self.paths['input_dir']) == os.path.realpath(self.paths['align_subdir']):
            errorState = True; errorText += "Error: Input and align directory may not be the same.\n"    
        
        if bool( errorText ):
            errorBox = QtGui.QMessageBox()
            errorBox.setText( errorText )
            errorBox.addButton( QtGui.QMessageBox.Ok )
            errorBox.setDefaultButton( QtGui.QMessageBox.Ok )
            errorBox.exec_()
        self.skulk.updatePaths( self.paths ) # I should really just have one paths in skulk
        return not errorState
        

        
    def updateAllViews( self, zorroObj = None ):
        for viewWidg in self.viewWidgetList:
            viewWidg.updateZorroObj( zorroObj = zorroObj )
        pass
    
#    def getCurrentImage( self ):
#        if self.currDMImage is None:
#            self.updateSelectedFile() # This fails safely if nothing is selected
#        return self.currDMImage
            
    def updateFileList( self, filelist=None ):
        print( "TODO: implement global update of all items in skulkManager" )
    
    
    def updateDict( self, dictname, key, funchandle, funcarg = None ):
        print( "updateDict: " + str(dictname) + " : " + str(key) + " : " + str(funchandle) + " : " + str(funcarg) )

        dictname[key] = funchandle()
        print( str(dictname[key] ) )
        if key == 'n_threads':
            for hostName, hostObj in self.skulk.hostInfo:
                hostObj.n_threads = funchandle()
            pass
        
        
    def updateZorroDefault( self, zorroAttrName, funcHandle, funcArg = None ):
        if isinstance( funcHandle, tuple):
            newVal = list( func() for func in funcHandle )
        else:
            newVal = funcHandle()
        
        print( "Changing zorroDefault."+zorroAttrName+ " from: " + 
                str(self.zorroDefault.__getattribute__(zorroAttrName)) + " to: " + str(newVal) )
                
        if newVal == 'none':
            newVal = None
        self.zorroDefault.__setattr__( zorroAttrName, newVal )
        # Stealing reference
        self.skulk.zorroDefault = self.zorroDefault


    def quitApp( self, event = None ):
        print( "Shutting down: " + str(event) )
        self.killSkulk()
        self.MainWindow.close()
        self.FileLocDialog.close()
        self.exit() 
        sys.exit( )
        
    def loadConfig( self, cfgfilename ):
        
        if cfgfilename is None:
            # open a dialog and ask user to pick a file
            cfgfilename = QtGui.QFileDialog.getOpenFileName( parent=self.MainWindow, caption="Load Config File", 
                                dir="", filter="Config files (*.ini)", selectedFilter="*.ini")[0]
        
        if cfgfilename == '':
            return
        else:
            self.cfgfilename = cfgfilename
        
        self.centralwidget.blockSignals(True)
        self.statusbar.showMessage( "Loading config file: " + self.cfgfilename  )
        config = ConfigParser.ConfigParser()
        config = ConfigParser.ConfigParser(allow_no_value = True)
        config.optionxform = str
        
        ##### Paths #####
        config.read( self.cfgfilename )
    
        try:
            self.paths['input_dir'] = config.get('automator_paths','input_dir')
            self.ui_FileLocDialog.leInputPath.setText( self.paths['input_dir'] )
            self.skulk.paths['input_dir'] = self.paths['input_dir']
        except: pass
        try:
            self.paths['output_dir'] = config.get('automator_paths','output_dir')
            self.ui_FileLocDialog.leOutputPath.setText( self.paths['output_dir'] )
            self.skulk.paths['output_dir'] = self.paths['output_dir']
        except: pass
        try:
            self.paths['raw_subdir'] = config.get('automator_paths','raw_subdir')
            self.ui_FileLocDialog.leRawPath.setText( self.paths['raw_subdir']  )
            self.skulk.paths['raw_subdir'] = self.paths['raw_subdir']
        except: pass
        try:
            self.paths['sum_subdir'] = config.get('automator_paths','sum_subdir')
            self.ui_FileLocDialog.leSumPath.setText( self.paths['sum_subdir']  )
            self.skulk.paths['sum_subdir'] = self.paths['sum_subdir']
        except: pass
        try:
            self.paths['align_subdir'] = config.get('automator_paths','align_subdir')
            self.ui_FileLocDialog.leAlignPath.setText( self.paths['align_subdir'] )
            self.skulk.paths['align_subdir'] = self.paths['align_subdir']
        except: pass
        try:
            self.paths['figure_subdir'] = config.get('automator_paths','figure_subdir')
            self.ui_FileLocDialog.leFiguresPath.setText( self.paths['figure_subdir'] )
            self.skulk.paths['figure_subdir'] = self.paths['figure_subdir']
            for viewWdgt in self.viewWidgetList:
                viewWdgt.viewCanvas.param['figurePath'] = self.paths['figure_subdir']
        except: pass

        try:
            self.paths['output_format'] = config.get('automator_paths','output_format')
            self.ui_FileLocDialog.comboOutputFormat.setCurrentIndex( self.ui_FileLocDialog.comboOutputFormat.findText(self.paths['output_format']) )
            self.skulk.paths['output_format'] = self.paths['output_format']
        except: pass
        
        ##### views #####
        for viewWidg in self.viewWidgetList:
            try:
                viewWidg.loadConfig( config )
                pass
            except: pass

        ###### Common configuration #####
        try:
            self.cfgCommon["registrator"] = config.get( 'automator_common', 'registrator' )
            self.comboRegistrator.setCurrentIndex( self.comboRegistrator.findText( self.cfgCommon["registrator"] ) )
        except: print( "Failed to set registrator: "  )
    
        ###### Cluster configuration #####
        try:
            self.cfgCluster["cluster_type"] = config.get( 'automator_cluster', 'cluster_type' )
            self.comboClusterType.setCurrentIndex( self.comboClusterType.findText( self.cfgCluster["cluster_type"] ) )
        except: print( "Failed to set cluster_type: "  )
        try:
            self.cfgCluster["n_processes"] = config.getint( 'automator_cluster', 'n_processes' )
            self.sbNProcesses.setValue( self.cfgCluster["n_processes"] )
        except: print( "Failed to set n_processes: "  )
        try:
            self.cfgCluster["n_threads"] = config.getint( 'automator_cluster', 'n_threads' )
            self.sbNThreads.setValue( self.cfgCluster["n_threads"] )
        except: print( "Failed to set n_processes: "  )
        
        ###### Gautomatch configuration #####
        
    
        # Load all the zorro parameters into zorroDefault
        self.zorroDefault.loadConfig( self.cfgfilename )
        # Update all the GUI elements
        
        try: self.comboTriMode.setCurrentIndex( self.comboTriMode.findText( self.zorroDefault.triMode ) )
        except: pass
        try: self.sbPeaksigThres.setValue( self.zorroDefault.peaksigThres )
        except: pass
        try: self.sbStartFrame.setValue( self.zorroDefault.startFrame )
        except: pass
        try: self.sbEndFrame.setValue( self.zorroDefault.endFrame )
        except: pass
        try: self.sbDiagWidth.setValue( self.zorroDefault.diagWidth )
        except: pass
        try: self.sbAutomax.setValue( self.zorroDefault.autoMax )
        except: pass
        try: self.cbSuppressOrigin.setChecked( self.zorroDefault.suppressOrigin )
        except: pass
        try: self.cbDoCTF.setChecked( self.zorroDefault.doCTF )
        except: pass
    
        try: self.comboCompressionExt.setCurrentIndex( self.comboCompressionExt.findText( self.zorroDefault.compress_ext ) )
        except: self.comboCompressionExt.setCurrentIndex( self.comboCompressionExt.findText( "none" ) )
        try: self.cbDoCompression.setChecked( self.zorroDefault.doCompression )
        except: pass
        try: self.cbSavePNG.setChecked( self.zorroDefault.savePNG )
        except: pass
        try: self.cbDoseFilter.setChecked( self.zorroDefault.doDoseFilter )
        except: pass
    
        try: # This is easier then blocking all the signals...
            shapePadded = np.copy( self.zorroDefault.shapePadded )
            self.sbShapePadX.setValue( shapePadded[1] )
            self.sbShapePadY.setValue( shapePadded[0] )
        except: print( "Failed to set sbShapePadX" )
        
        try: 
            fouCrop = np.copy( self.zorroDefault.fouCrop )
            self.sbFouCropX.setValue( fouCrop[1] )
            self.sbFouCropY.setValue( fouCrop[0] )
        except: pass
        
        try: self.sbPixelsize.setValue( self.zorroDefault.pixelsize )
        except: pass
        try: self.sbVoltage.setValue( self.zorroDefault.voltage )
        except: pass
        try: self.sbC3.setValue( self.zorroDefault.C3 )
        except: pass
        try: self.sbMaxShift.setValue( self.zorroDefault.maxShift )
        except: pass
        try: self.comboOriginMode.setCurrentIndex( self.comboOriginMode.findText( self.zorroDefault.originMode ) )
        except: pass
        try: self.cbPreshift.setChecked( self.zorroDefault.preShift )
        except: pass
        try: self.cbSaveC.setChecked( self.zorroDefault.saveC )
        except: pass
        try: self.comboBmode.setCurrentIndex( self.comboBmode.findText( self.zorroDefault.Bmode ) )
        except: pass
        try: self.sbBrad.setValue( self.zorroDefault.Brad )
        except: pass
        try: self.comboWeightMode.setCurrentIndex( self.comboWeightMode.findText( self.zorroDefault.weightMode ) )
        except: pass
        try: self.sbSubpixReg.setValue( self.zorroDefault.subPixReg )
        except: pass
        try: self.comboShiftMethod.setCurrentIndex( self.comboShiftMethod.findText( self.zorroDefault.shiftMethod ) )
        except: pass
        
    
        try: self.sbShapePadY.setValue( self.zorroDefault.shapePadded[0] )
        except: print( "Failed to set sbShapePadY" )
        try: self.sbFouCropY.setValue( self.zorroDefault.fouCrop[0] )
        except: pass
        try: self.leCachePath.setText( self.zorroDefault.cachePath )
        except: pass
        try: self.comboFFTWEffort.setCurrentIndex( self.comboFFTWEffort.findText( self.zorroDefault.fftw_effort ) )
        except: pass
        
        try: self.sbNThreads.setValue( self.zorroDefault.n_threads )
        except: pass
    
        # Gautomatch
        try: 
            self.cfgGauto['boxsize'] = config.get( 'automator_gauto', 'boxsize' )
            self.leGautoBoxsize.setText( self.cfgGauto['boxsize'] )
        except: pass
        try: 
            self.cfgGauto['diameter'] = config.get( 'automator_gauto', 'diameter' )
            self.leGautoDiameter.setText( self.cfgGauto['diameter'] )
        except: pass
        try: 
            self.cfgGauto['min_dist'] = config.get( 'automator_gauto', 'min_dist' )
            self.leGautoMin_Dist.setText( self.cfgGauto['min_dist'] )
        except: pass
        try: 
            self.cfgGauto['T'] = config.get( 'automator_gauto', 'T' )
            self.leGautoTemplates.setText( self.cfgGauto['T'] )
        except: pass
        try: 
            self.cfgGauto['ang_step'] = config.get( 'automator_gauto', 'ang_step' )
            self.leGautoAng_Step.setText( self.cfgGauto['ang_step'] )
        except: pass
        try: 
            self.cfgGauto['speed'] = config.get( 'automator_gauto', 'speed' )
            self.leGautoSpeed.setText( self.cfgGauto['speed'] )
        except: pass
        try: 
            self.cfgGauto['cc_cutoff'] = config.get( 'automator_gauto', 'cc_cutoff' )
            self.leGautoCCCutoff.setText( self.cfgGauto['cc_cutoff'] )
        except: pass
        try: 
            self.cfgGauto['lsigma_D'] = config.get( 'automator_gauto', 'lsigma_D' )
            self.leGautoLsigma_D.setText( self.cfgGauto['lsigma_D'] )
        except: pass
        try: 
            self.cfgGauto['lsigma_cutoff'] = config.get( 'automator_gauto', 'lsigma_cutoff' )
            self.leGautoLsigma_Cutoff.setText( self.cfgGauto['lsigma_cutoff'] )
        except: pass
        try: 
            self.cfgGauto['lave_D'] = config.get( 'automator_gauto', 'lave_D' )
            self.leGautoLave_D.setText( self.cfgGauto['lave_D'] )
        except: pass
        try: 
            self.cfgGauto['lave_min'] = config.get( 'automator_gauto', 'lave_min' )
            self.leGautoLave_Min.setText( self.cfgGauto['lave_min'] )
        except: pass
        try: 
            self.cfgGauto['lave_max'] = config.get( 'automator_gauto', 'lave_max' )
            self.leGautoLave_Max.setText( self.cfgGauto['lave_max'] )
        except: pass
        try: 
            self.cfgGauto['lp'] = config.get( 'automator_gauto', 'lp' )
            self.leGautoLP.setText( self.cfgGauto['lp'] )
        except: pass
        try: 
            self.cfgGauto['hp'] = config.get( 'automator_gauto', 'hp' )
            self.leGautoHP.setText( self.cfgGauto['hp'] )
        except: pass
        try: 
            self.cfgGauto['pre_lp'] = config.get( 'automator_gauto', 'pre_lp' )
            self.leGautoLPPre.setText( self.cfgGauto['pre_lp'] )
        except: pass
        try: 
            self.cfgGauto['pre_hp'] = config.get( 'automator_gauto', 'pre_hp' )
            self.leGautoHPPre.setText( self.cfgGauto['pre_hp'] )
        except: pass
        # Plotting for Gautomatch
        try: 
            self.cfgGplot['do_pre_filter'] = config.getboolean( 'automator_gplot', 'do_pre_filter' )
            self.cbGautoDoprefilter.setChecked( self.cfgGplot['do_pre_filter'] )
        except: pass
        try: 
            self.cfgGplot['write_ccmax_mic'] = config.getboolean( 'automator_gplot', 'write_ccmax_mic' )
            self.cbGautoPlotCCMax.setChecked( self.cfgGplot['write_ccmax_mic'] )
        except: pass
        try: 
            self.cfgGplot["write_pref_mic"] = config.getboolean( 'automator_gplot', "write_pref_mic" )
            self.cbGautoPlotPref.setChecked( self.cfgGplot["write_pref_mic"] )
        except: pass
        try: 
            self.cfgGplot["write_bg_mic"] = config.getboolean( 'automator_gplot', "write_bg_mic" )
            self.cbGautoPlotBG.setChecked( self.cfgGplot["write_bg_mic"] )
        except: pass
        try: 
            self.cfgGplot["write_bgfree_mic"] = config.getboolean( 'automator_gplot', "write_bgfree_mic" )
            self.cbGautoPlotBGFree.setChecked( self.cfgGplot["write_bgfree_mic"] )
        except: pass
        try: 
            self.cfgGplot["write_lsigma_mic"] = config.getboolean( 'automator_gplot', "write_lsigma_mic" )
            self.cbGautoPlotLsigmaFree.setChecked( self.cfgGplot["write_lsigma_mic"] )
        except: pass
        try: 
            self.cfgGplot["write_mic_mask"] = config.getboolean( 'automator_gplot', "write_mic_mask" )
            self.cbGautoPlotMaskFree.setChecked( self.cfgGplot["write_mic_mask"] )
        except: pass
    
        self.centralwidget.blockSignals(False)
        # End of Automator.loadConfig
    
    def saveConfig( self, cfgfilename ):
        if cfgfilename is None:
            cfgfilename = QtGui.QFileDialog.getSaveFileName(
                parent=self.MainWindow,caption="Save Config File", dir="", filter="Config files (*.ini)", 
                selectedFilter="*.ini")[0]
            
        if cfgfilename == '':
            return
        else:
            # Force extension to .ini
            cfgfilename = os.path.splitext( cfgfilename )[0] + ".ini"
            self.cfgfilename = cfgfilename
            
        self.statusbar.showMessage( "Saving configuration: " + cfgfilename )
            
#        # Overwrite the file if it already exists

        self.zorroDefault.saveConfig( cfgfilename )
        # Read in the config prepared by Zorro
        config = ConfigParser.ConfigParser()
        config = ConfigParser.ConfigParser(allow_no_value = True)
        config.optionxform = str
        
        config.read( cfgfilename )
        ### PATH ###
        config.add_section('automator_paths')
        for key in self.paths:
            config.set( 'automator_paths', key, self.paths[key] )
            
        config.add_section('automator_common')
        for key in self.cfgCommon:
            config.set( 'automator_common', key, self.cfgCommon[key] )
    
        config.add_section('automator_cluster')
        for key in self.cfgCluster:
            config.set( 'automator_cluster', key, self.cfgCluster[key] )
            
        config.add_section('automator_gauto')
        for key in self.cfgGauto:
            config.set( 'automator_gauto', key, self.cfgGauto[key] )
            
        config.add_section('automator_gplot')
        for key in self.cfgGplot:
            config.set( 'automator_gplot', key, self.cfgGplot[key] )   
            
        # Add viewWidgets using their built-in config writers
        for viewWidg in self.viewWidgetList:
            viewWidg.saveConfig( config )
            
        try:
            # Open in append mode
            cfgfh = open( self.cfgfilename, 'ab' )
            config.write( cfgfh )
            cfgfh.close()
        except:
            print( "Error in loading config file: " + self.cfgfilename )
            

    
    def openPathDialog( self, pathname, openDialog ):
        # Comments on how to handle updates from a lineedit Qt object
        # http://stackoverflow.com/questions/12182133/pyqt4-combine-textchanged-and-editingfinished-for-qlineedit
        
        if bool(openDialog):
            pathDialog = QtGui.QFileDialog()
            pathDialog.setFileMode( QtGui.QFileDialog.Directory )
            pathDialog.setOption( QtGui.QFileDialog.ShowDirsOnly, True )
            self.paths[pathname] = str(pathDialog.getExistingDirectory(self.MainWindow,pathname, ""))
            print( "New path for "+ pathname + " : " + self.paths[pathname] )
            self.skulk.updatePaths( self.paths )
        else:
            # Get directory from the lineedit object
            print( "TODO" )
            
        #self.FileLocDialog.raise_()
        self.FileLocDialog.activateWindow()

        if pathname == 'input_dir':
            self.ui_FileLocDialog.leInputPath.setText( self.paths[pathname] )
        elif pathname == 'output_dir':
            self.ui_FileLocDialog.leOutputPath.setText( self.paths[pathname] )
        elif pathname == 'raw_subdir': 
            # Make a relative path to output, TODO, make this smarter
            #self.paths[pathname] = os.path.relpath( self.paths[pathname], self.paths['output_dir'] )
            self.ui_FileLocDialog.leRawPath.setText( self.paths[pathname] )
        elif pathname == 'sum_subdir':
            # Make a relative path to output 
            #self.paths[pathname] = os.path.relpath( self.paths[pathname], self.paths['output_dir'] )
            self.ui_FileLocDialog.leSumPath.setText( self.paths[pathname] )
        elif pathname == 'align_subdir':
            # Make a relative path to output
            #self.paths[pathname] = os.path.relpath( self.paths[pathname], self.paths['output_dir'] )
            self.ui_FileLocDialog.leAlignPath.setText( self.paths[pathname] )
        elif pathname == 'figure_subdir':
            #self.paths[pathname] = os.path.relpath( self.paths[pathname], self.paths['output_dir'] )
            self.ui_FileLocDialog.leFiguresPath.setText( self.paths[pathname] )
        pass
        self.skulk.updatePaths( self.paths )
    
    
def main():
    try:
        mainGui = Automator()
    except SystemExit:
        sys.exit()
        exit
        
# Instantiate a class
if __name__ == '__main__': # Windows multiprocessing safety
    main()
