# -*- coding: utf-8 -*-
"""
Zorro Skulk manager, is a low-overhead queue manager for Zorro.

Basic intention is to manage many zorro objects, submit to queues.

TODO: may want to track meta-data over entire data set for eventually maximum-likelihood approach
Such as having the correlations for the entire set.

Created on Thu Dec 24 22:31:16 2015
@author: Robert A. McLeod
@email: robbmcleod@gmail.com OR robert.mcleod@unibas.ch
"""

import numpy as np
import zorro
# TO DO: multi-process, or multi-thread with subprocess?  multiprocessing module is quite buggy in my experience.
# We need multiprocessing for a Queue
# Always protect mp with if __name__ == "__main__" for Windows
# Multiprocessing is really a horrible little module
#import multiprocessing as mp
# import threading, subprocess
import subprocess as sp
import Queue
from copy import deepcopy
import time
import os, os.path
import glob

try: 
    from PySide.QtCore import Signal
    from PySide.QtCore import QThread, QMutex
except:
    print( "ZorroSkulkManager failed to load PySide; no communication with Automator possible" )


verbose = 0

#################### SKULKHOST CLASS ###############################
class skulkHost(object):
    
    def __init__(self, hostname, workerFunction, parentMessenger, 
                 n_threads = None, cachepath = None, qsub_script=None ):
        self.hostName = hostname
        self.parentMessenger = parentMessenger
        self.zorroObject = None
        
        # CANNOT get an instantiated object here so we cannot pass in bound function handles directly
        if workerFunction == 'local':
            self.workerFunction = self.submitLocalJob
        elif workerFunction == 'dummy':
            self.workerFunction = self.submitDummyJob
        else:
            self.workerFunction = workerFunction
            
        self.subprocess = None
        self.n_threads = n_threads
        self.cachePath = cachepath
        self.qsub_script = qsub_script
    
        
    def kill(self):
        # This could get a little tricky based on whatever workerFunction we are using.  Maybe I should 
        # subclass skulkHost instead.
        try:
            self.subprocess.kill()
        except: pass
        
    def queryHost(self):
        # Would it be better to have some sort of event driven implementation?
        if self.subprocess == None:
            return 1 # Free
        poll = self.subprocess.poll()
        if poll == None: # Still working
            return 0
        else:
            self.returnZorroObj() # Try
            return 1 # Free
            
            
    def returnZorroObj(self):
        if self.zorroObject != None:
            # Error handling done by the manager
            self.parentMessenger.put( [self.zorroObject.files['config'], 'fini'] )
            self.zorroObject = None
    
    def submitLocalJob( self, zorroObj ):
        # Local job, no queue
        # Check/copy that stackName exists in the right place
        self.zorroObject = zorroObj
        
        self.zorroObject.n_threads = self.n_threads
        if self.cachePath != None: # Force use of host's cachePath if it's present...
            self.zorroObject.cachePath = self.cachePath
        # self.zorroObject.verbose = 0 # For debugging verbose is set globally at the top of the file
        self.zorroObject.METAstatus = 'proc'
        
        # Execute zorro
        self.zorroObject.saveConfig()
        # zorro_script = zorro.util.which( 'zorro.py' )
        zorro_script = os.path.join( os.path.split( zorro.__file__ )[0], "zorro.py" )
        
        commandStr = "python " + zorro_script + " -c " + self.zorroObject.files['config']

        print( "Subprocess local exec: " + commandStr )
        # Seems that shell=True is more or less required.
        self.subprocess = sp.Popen( commandStr, shell=True )
        
    
    def submitDummyJob( self, zorroObj ):
        self.zorroObject = zorroObj
        # Cheap job for testing purposes
        time.sleep( 5 )

        pass
    
    def submitJob_viaSSH( self, zorroObj ):
        # Remote job, no queue#################### SUBPROCESS FUNCTIONS ###############################
        print( "Submission of jobs via SSH is not implemented yet." )
        pass
    
    def qsubJob( self, zorroObj ):
        # Local queue, i.e. on a login node
        if self.qsub_script == None:
            print( "Setup qsub_script before submitting jobs via qsub!" )
            return
        self.zorroObject = zorroObj
    
        self.zorroObject.n_threads = self.n_threads
        if self.cachePath != None: # Force use of host's cachePath if it's present...
            self.zorroObject.cachePath = self.cachePath
        # zorroObject.verbose = 0

        
        # Execute zorro
        self.zorroObject.saveConfig()
        
        # Write a configuration file
        print( "TODO: Write a configuration file" )
        # Write a qsub.bash script
        print( "TODO: Write a qsub.bash script" )
        # Qsub host is always free to accept new jobs
        # self.parentMessenger.put( (self.hostName, 'ready') )
        # How to tell if the job is finished?  Launch another process to look for it?
        pass
    
    def qsubJob_visSSH( self, zorroObject ):
        # Remote queue, submit via SSH needs some sort of distributed file system?  Or can we use rsync (if 
        # this is to be used as a daemon in this case)
        print( "Submission of jobs via SSH is not implemented yet." )
        pass

    pass


#################### SKULKPRIORITYQUEUE CLASS ###############################
class skulkPriorityQueue(object):
    """
    This class manages all the files the watcher in skulkManager finds, and prioritizes them as desired 
    by the user.  It is otherwise actually a stack, not a queue, as the latest files are processed first.
    """
    def __init__(self):
        #self.__mutex = threading.Lock()
        self.__mutex = QMutex()
        self.__iterIndex = 0
        
        # Each item in the list contains a tuple of: 
        # {'stackName': ( mtime, bytesize, status, userPriority, zorroObject )
        self.__queue = [] # List of zorroObjects
        
        pass
 
    def put( self, zorroObject ):
        self.__mutex.lock()
        if zorroObject.METAstatus == 'ready':
            priority = zorroObject.METApriority / zorroObject.METAmtime
        else:
            priority = -1.0
        try:
            # Add to queue
            self.__queue.append( [priority, zorroObject] )
            # Re-sort on priority
            self.__queue.sort()
        except:
            print( "Something bad (likely QThread related) happened in skulkPriorityQueue::put" )
        self.__mutex.unlock()
            
    def __len__(self):
        self.__mutex.lock()
        length = len( self.__queue )
        self.__mutex.unlock()
        return length
        
    def __getitem__( self, key ):
        self.__mutex.lock()
        if isinstance(key,int):
            # self.__queue.sort()
            self.__mutex.unlock()
            return self.__queue[key][1]
        elif isinstance( key, str ):
            self.__mutex.unlock()
            return self.getByName( key )
            
    def isLocked( self ):
        state = self.__mutex.tryLock()
        if state:
            self.__mutex.unlock()
        return not state
        
    # Thread-safe iteration with yield seems impossible so it was removed
        
    def getNextReady( self ):
        self.__mutex.lock()
        for priority, zorroObject in self.__queue:
            if zorroObject.METAstatus == 'stable':
                self.__mutex.unlock()
                return zorroObject
        # If we get here, there's no ready stacks
        self.__mutex.unlock()
        return None
        
    def updatePriority( self, key, newPriority ):
        """
        High priorty jobs will run first.
        
        -1.0 is used for jobs that aren't to be run yet.
        """
        self.__mutex.lock()
        if isinstance(key,int):
            self.__queue[key][0] = newPriority
        elif isinstance( key, str ):
            for J, (priority, zorroObj) in enumerate(self.__queue):
                if zorroObj.files['stack'] == key:
                    self.__queue[J][0] = newPriority
        self.__mutex.unlock()
        
    def keys(self):
        self.__mutex.lock()
        keysList = [""] * len(self.__queue )
        for J, (priority, zorroObject) in enumerate( self.__queue ):
            keysList[J] = zorroObject.files['stack']
        self.__mutex.unlock()
        return keysList
            
    def getByName( self, key ):
        self.__mutex.lock()
        for priority, zorroObject in self.__queue:
            if zorroObject.files['stack'] == key:
                self.__mutex.unlock()
                return zorroObject
        self.__mutex.unlock()
        return None
        
    def __delitem__( self, key ):
        self.__mutex.lock()
        for priority, zorroObject in self.__queue:
            if zorroObject.files['stack'] == key:
                self.__queue.remove( [priority,zorroObject] )
                break
        self.__mutex.unlock()
        
    def remove( self, key ):
        # NO MUTEX, __delitem__ TAKES MUTEX
        self.__delitem__( key )
        

#################### SKULKMANAGER CLASS ###############################
class skulkManager(QThread):
    
    # Signals must be statically declared?
    try:
        automatorSignal = Signal( str, str )
    except:
        automatorSignal = None
    
    def __init__(self, inputQueue=None ):
        QThread.__init__(self)
        
        # Processible file extensions
        self.globPattern = ['*.dm4', '*.dm3', '*.mrc', '*.mrcs', '*.hdf5', '*.h5', 
                 '*.dm4.bz2', '*.dm3.bz2', '*.mrc.bz2', '*.mrcs.bz2', '*.hdf5.bz2', '*.h5.bz2', 
                 '*.dm4.gz', '*.dm3.gz', '*.mrc.gz', '*.mrcs.gz', '*.hdf5.gz', '*.h5.gz',
                 '*.dm4.zip', '*.dm3.zip', '*.mrc.zip', '*.mrcs.zip', '*.hdf5.zip', '*.h5.zip', 
                 '*.dm4.7z', '*.dm3.7z', '*.mrc.7z', '*.mrcs.7z', '*.hdf5.7z', '*.h5.7z']
        # We need to insert dummy zorros into the queue if the file is compressed because it will be visible 
        # off disk?  Or can we change unzip to direct to the cache???
        
        self.debug = 3
                
        # Default object for copying and base level actions        
        self.zorroDefault = zorro.ImageRegistrator()
        
        # A dict of completed zorro objects that uses the stackname as a key
        self.skulkFini = {}
        
        # Mutex-protected queue of the image stacks to be worked on
        self.skulkQueue = skulkPriorityQueue()
        self.subproccesses = {}
        # Pool is necessary to start asynchronous processes

        # Messages from the hosts to the skulkManager (i.e. I've finished, or, I've fallen and I can't get up again)
        # This is unidirectional from the host to the skulk manager
        # self.parentAutomator = None # This seems not thread-safe?
        self.hostMessageQueue = Queue.Queue(maxsize=0) 
        
        # Servers that will accept jobs, is a dict of skulkHosts
        self.hostInfo = {}
        
        # Hosts need name, IP, n_threads, FFTW_effort, fftw_wisdom location, cachePath
        self.useQueue = False
        # Ideally I should be able to generate the submission script on the fly, and just call qsub
        self.qsubTemplate = "somescript.bash" # Basqsubh, python, whatever, just needs to call script appropriately
        
        # FILE WATCHER
        self.paths = { "input_dir":"", "output_dir":"", "raw_subdir":"", "sum_subdir":"", 
                      "align_subdir":"", "figure_subdir":"" }

        # TODO: Do we want Paramiko for running remote jobs via SSH?  Or restrict ourselves to qsub locally?
        # https://github.com/paramiko/paramiko
        # May want it regardless if we don't want to run automator on the cluster computer via X
        # It's not in Anaconda by default, but it's in both conda and pip
        
        self.__STOP_POLLING = False
        # bytes, should be bigger than Digital Micrograph header size, but smaller than any conceivable image
        self.__MIN_FILE_SIZE = 1048576 
        self.sleepTime = 1.0
        pass
    
    def run(self):
        loopCounter = 0
        print( "skulk.__STOP_POLLING = " + str(self.__STOP_POLLING))
        while not self.__STOP_POLLING:
            # Check queues if they have returned a result
            # This doesn't work how I expect...
            while self.hostMessageQueue.qsize() > 0:
                message = self.hostMessageQueue.get()
                print( "Received message: " + str(message) )
                # Process the message
                if message[1]  == 'fini':
                    logName = message[0]
                    print( "######## ZorroObj "+ logName + " RETURNED " )
                    
                     # Reload the config file so we have all the information for plotting
                    # and add it to the list of finished stacks
                    newZorroObj = zorro.ImageRegistrator()
                    newZorroObj.loadConfig( logName )
                    
                    if newZorroObj.METAstatus != 'fini':
                        print( "ERROR: skulkManager returned zorro config is not finished" )
                    
                    # Remove the original key from the queue
                    print( "A: " + str( self.skulkQueue.isLocked() ) )
                    self.skulkQueue.remove( newZorroObj.files['original'] )
                    if self.automatorSignal != None:
                        print( "##### REMOVING FROM AUTOMATOR: " + newZorroObj.files['original']  )
                        self.autoUpdate( newZorroObj.files['original'], 'remove' )
                        

                    self.skulkFini[newZorroObj.files['config']] = newZorroObj
                    if self.automatorSignal != None:
                        print( "##### ADDING TO AUTOMATOR: " + newZorroObj.files['config']  )
                        self.autoUpdate( newZorroObj.files['config'], 'fini' )

                else:
                    print( "TODO: skulkManager::mainPollingLoop: Process this message : " + str( message ) )
            
            freeHosts = []
            statusMessage = "%s "%loopCounter
            for J, hostKey in enumerate( self.hostInfo ):
                poll = self.hostInfo[hostKey].queryHost()
                if poll == 1:
                    freeHosts.append( self.hostInfo[hostKey] )
                statusMessage += " | %s"%hostKey + ": %s"%poll
                    
            print( statusMessage )
            
            print( "G: " + str( self.skulkQueue.isLocked() ) )
            if len( freeHosts ) > 0 and len( self.skulkQueue ) > 0:
                
                # Dispatch jobs to newly freed queues
                for J, host in enumerate( freeHosts ):
                    print( "H: " + str( self.skulkQueue.isLocked() ) )
                    zorroObjToProcess = self.skulkQueue.getNextReady()
                    
                    if zorroObjToProcess != None:
                        print( "Dispatching : %s"%zorroObjToProcess.files['stack'] + " to: %s"%host.workerFunction.__name__ )
                        
                        # TODO: if workerFunction fails, execute some nice error handling on the host
                        host.workerFunction( zorroObjToProcess )                 

                        zorroObjToProcess.METAstatus = 'proc'

                        if self.automatorSignal != None:
                            self.autoUpdate( zorroObjToProcess.files['stack'], 'proc' )

                        print( "I " + str( self.skulkQueue.isLocked() ) )
                        self.skulkQueue.updatePriority( zorroObjToProcess.files['stack'], -1.0 )
            
            # Now check the watch directory, populates self.skulkInfo
            self.pollWatcherDirectory()
            
            # Sleep
            loopCounter += 1
            self.sleep( self.sleepTime )

    
        
    
    def autoUpdate( self, key, message ):
        self.automatorSignal.emit( key, message )
    
    def __getitem__( self, key ):
        # Search through skulkFini first, then skulkQueue
        if self.skulkFini.has_key(key):
            return self.skulkFini[key]
        else: 
            print( "C: " + str( self.skulkQueue.isLocked() ) )
            return self.skulkQueue.getByName( key ) # Will return None rather than raising a KeyError
            
    def __len__(self):
        print( "D: " + str( self.skulkQueue.isLocked() ) )
        return len(self.skulkFini) + len(self.skulkQueue)
            
    def keys( self ):
        # Return all object names in skulkFini + skulkQueue
        print( "E: " + str( self.skulkQueue.isLocked() ) )
        zorroKeys = self.skulkQueue.keys()
        zorroKeys.extend( self.skulkFini.keys() )
        return zorroKeys
    
    def remove( self, key ):
        """
        WARNING: this function deletes all files associated with a key, not just the log.
        """
        print( "FINI\n" + str(self.skulkFini) )
        print( "QUEUE|n" + str(self.skulkQueue) )
        try:
            print( "F: " + str( self.skulkQueue.isLocked() ) )
            deleteZorro = self.skulkQueue.getByName( key )
            if deleteZorro != None:
                self.skulkQueue.remove( key )
            print( "removed from Queue: " + key + " as " + str(deleteZorro) )
        except:
            try: 
                deleteZorro = self.skulkFini.pop(key)
                print( "removed from Finish list: " + key + " as " + str(deleteZorro) )
            except: 
                print( "Couldn't find key: " + str(key) )
                return
        
        if key.endswith( '.log' ) or (deleteZorro != None and 'config' in deleteZorro.files):
            try:
                deleteZorro = zorro.ImageRegistrator()
                deleteZorro.loadConfig( key )
                # DELETE ALL FILES
                currDir = os.path.realpath('.')
                os.chdir( os.path.split(key)[0] ) # Change directory so relative paths are valid
                for filename in deleteZorro.files.values():
                    try: os.remove( filename )
                    except: pass
                os.chdir( currDir )
            except: # Failed to delete
                pass
        else: # We can't find a log with a list of files so we can only remove the key
            try: os.remove( key )
            except: pass
        # politely recommend garbage collection
        del deleteZorro
        
    def reprocess( self, key ):
        """
        Reprocesses a selected file with current conditions, and also gives the file extra-high priority so 
        it goes to top of the stack.
        """
        # If it's finished, move it back into the queue
        if key in self.skulkFini:
            zorroObj = self.skulkFini.pop( key )
            # Just delete the log, everything else should be over-written
            os.remove( zorroObj.files['config'] )
            print( "F2: " + str( self.skulkQueue.isLocked() ) )
        else: # still in the Queue
            # Unfinished, potentially unknown current status.  Delete everything but the raw stack and start 
            # again.
            self.skulkQueue.remove( key )
            # TODO: delete intermediates if found?
            
        print( "F3: " + str( self.skulkQueue.isLocked() ) )
        if os.path.isabs( zorroObj.files['stack'] ):
             newKey = zorroObj.files['stack']
        else:
            newKey = os.path.join( os.path.split( zorroObj.files['config'] )[0], zorroObj.files['stack'] )
            
        self.buildNewZorroObj( newKey )
        # Try to give it extra-high priority
        self.skulkQueue.updatePriority( newKey, 1E10 )
        self.autoUpdate( newKey, 'new' )

        
    def __get_item__(self, key ):
        if self.skulkFini.has_key( key ):
            return self.skulkFini[key]
        print( "J: " + str( self.skulkQueue.isLocked() ) )
        return self.skulkQueue[key] # Can return None
        
    def kill(self):
        self.__STOP_POLLING = True
        for hostObj in self.hostInfo.values():
            hostObj.kill()
        # Go through the skulkQueue and change any 'proc' statuses to 'stable'
        print( "K1: " + str(self.skulkQueue.isLocked()) )
        if not self.skulkQueue.isLocked():
            print( "Forcing unlock of skulkQueue" )
            self.skulkQueue._skulkPriorityQueue__mutex.unlock()
            
        for J in np.arange( len( self.skulkQueue ) ):
            zorroObj = self.skulkQueue[J]
            if zorroObj.METAstatus == 'proc':
                zorroObj.METAstatus = 'stable' 
        print( "K2: " + str( self.skulkQueue.isLocked() ) )       
        print( "TODO: check if host moved original stack files before kill order." )
        
    def initHosts( self ):
        # TODO: this is currently controlled by automator
        pass
     
    def updatePaths( self, newPaths ):
        for key, value in newPaths.items():
            self.paths[key] = value
        # figure out the relative paths to the config log file, which goes to output_dir
        self.paths['raw_subdir'] = os.path.relpath( self.paths['raw_subdir'], start=self.paths['output_dir'] )
        self.paths['sum_subdir'] = os.path.relpath( self.paths['sum_subdir'], start=self.paths['output_dir'] )
        self.paths['align_subdir'] = os.path.relpath( self.paths['align_subdir'], start=self.paths['output_dir'] )
        self.paths['figure_subdir'] = os.path.relpath( self.paths['figure_subdir'], start=self.paths['output_dir'] )
            
    def inspectOutputDirectory(self):
        """
        Check the output directory for logs.  In general this is only called when a start is issued.
        """
        logList = glob.glob( os.path.join( self.paths['output_dir'], "*.log" ) )
        for log in logList:
            if log.endswith( "ctffind3.log" ): # Avoid adding CTF logs if they happen to be in the same place.
                continue
            
            newZorroObj = zorro.ImageRegistrator()
            newZorroObj.loadConfig( log )
            if newZorroObj.METAstatus == 'fini':
                self.skulkFini[ newZorroObj.files['config'] ] = newZorroObj
                if self.automatorSignal != None:
                    self.autoUpdate( newZorroObj.files['config'], newZorroObj.METAstatus )

            else: # Suggest re-run
                self.buildNewZorroObj( newZorroObj.files['stack'], status='new' )
                if self.automatorSignal != None:
                    self.autoUpdate( newZorroObj.files['stack'], 'new' )


        
    def pollWatcherDirectory(self):
        """
        Poll the input directory for new stack files, and dispatch them to the hosts as they are free.
        """
        # And for every existing key name, check if it's updated.
        print( "L: " + str( self.skulkQueue.isLocked() ) )
        for J in np.arange( len( self.skulkQueue ) ):
            zorroObj = self.skulkQueue[J]
            # It is a massive pain to tell if a file is open in another program or 'finished'.
            try:
                # 2.) if it has a newer mtime OR
                # Mmm, Gatan DM opens the file for buffering and takes awhile to start writing to it...
                # We could say the file has to be a certain size before we consider it?
                
                newMTime = os.path.getmtime( zorroObj.files['stack'] )
                newSize = os.path.getsize( zorroObj.files['stack'] )
                if (newMTime > zorroObj.METAmtime
                        or newSize > zorroObj.METAsize 
                        or newSize < self.__MIN_FILE_SIZE 
                        or zorroObj.METAstatus == 'new'):
                    # I guess there can be problems here if there's a lot of network congestion?  
                    # We can at least force files to go from new -> changing -> stable, that could help a bit.
                        
                    if self.debug >= 3: 
                        print( "Changing file: " + zorroObj.files['stack'] + ", mtime = %.1f"%newMTime +", size = " +str(newSize/1.0E6) +" MB" )
                        
                    zorroObj.METAmtime = newMTime
                    zorroObj.METAsize = newSize
                    zorroObj.METAstatus = 'changing'
                elif zorroObj.METAstatus == 'changing':
                    if self.debug >= 3: 
                        print( "Stable file: " + zorroObj.files['stack'] + ", mtime = %.1f"%newMTime +", size = " +str(newSize/1.0E6)+" MB" )
                    # 3.) If mtime and/or size was changing but now it's stable
                    zorroObj.METAstatus = 'stable'
                    #print( "Debug 1" )
                    zorroObj.METApriority = np.minimum( zorroObj.METApriority, 1.0 )
                    #print( "Debug 2" )
                    # try:
                    # if self.parentAutomator != None:
                    if self.automatorSignal != None:
                        #print( "Debug enter E" )
                        # self.parentAutomator.updateFromSkulk( zorroObj.files['stack'], 'stable' )
                        self.autoUpdate( zorroObj.files['stack'], 'stable' )
                        #print( "Debug exit E" )
                    #print( "Debug 3" )
                    #              except: pass
            
            except os.error: #os.error, if stackName doesn't exist
                # We often move the file and it's subprocessed, so how to know if the process is still on-going?
            
                if zorroObj.METAstatus == 'proc':
                    pass # Do nothing, we move files about during processing
                elif zorroObj.METAstatus == 'fini':
                    print( "FIXME: zorroObject " + zorroObj.files['stack'] + " is finished." )
                else:
                    # if self.debug >= 2: 
                    print( "File deleted from disk?: " + zorroObj.files['stack'] )
                    # 4.) except: was deleted, moved, renamed?
                    # try:
                    if self.automatorSignal != None:
                        #print( "Debug enter F" )
                        self.autoUpdate( zorroObj.files['stack'], 'remove' )
                        # self.parentAutomator.updateFromSkulk( zorroObj.files['stack'], 'remove' )
                        #print( "Debug exit F" )
                    #except: pass
                    print( "Mutex status R: " + str(self.skulkQueue.isLocked() ) )
                    self.skulkQueue.remove( zorroObj.files['stack'] )
        #print( "Debug 5" )            
        newGlobList = []
        for globPat in self.globPattern:
            newGlobList.extend( glob.glob( os.path.join( self.paths['input_dir'], globPat ) ) )
        for J, globFile in enumerate(newGlobList):
            # This is a little dangerous for overflow problems if the glob pattern doesn't match the keyed IDs
            newGlobList[J] = os.path.realpath( globFile )
            

        # Only let new files in
        print( "S: " + str( self.skulkQueue.isLocked() ) )
        newGlobList = list( set(newGlobList) - set(self.skulkQueue.keys()) )
        # For some reason this set operation isn't working
        
        
        for globItem in newGlobList:
            if self.debug >= 2: 
                print( "New file: " + globItem )
                self.buildNewZorroObj( globItem )
            
        pass
    
    def buildNewZorroObj(self, filename, status='new' ):
        newZorroObj = deepcopy( self.zorroDefault )
        
        # Populate names, both the inputs and the config file are real paths by convention.  The various 
        # sub-outputs are relative to the config file. 
        newZorroObj.files['stack'] = os.path.realpath( filename )
        newZorroObj.files['original'] = newZorroObj.files['stack']
        baseName = os.path.splitext( os.path.basename( filename ) )[0]
        
        # We strip the compress_ext if it's present from the logname
        # The log file is always a real-path, so the others can be relative paths.  This allows the user 
        # to move the whole structure later.
        newZorroObj.files['config'] = os.path.join( os.path.realpath(self.paths['output_dir']), 
            os.path.basename( filename.rstrip(newZorroObj.compress_ext) ) + ".log" )
            
        # These should all be relative to config if the updatePaths function was called
        newZorroObj.files['sum'] = os.path.join( self.paths['sum_subdir'], baseName + "_zorro.mrc" )
        newZorroObj.files['align'] = os.path.join( self.paths['align_subdir'], baseName + "_zorro_movie.mrcs" )
        newZorroObj.files['figurePath'] = self.paths['figure_subdir']
        newZorroObj.files['moveRawPath'] = self.paths['raw_subdir']
        
        newZorroObj.verbose = verbose
        newZorroObj.METAmtime = os.path.getmtime(filename)
        newZorroObj.METAsize = os.path.getsize(filename)
        newZorroObj.METAstatus = status
        newZorroObj.METApriority = -1.0

        self.skulkQueue.put( newZorroObj )
        self.autoUpdate( newZorroObj.files['stack'], status )

if __name__ == "__main__":
#    # Queue example:    
#    
#    unzipFiles = glob.glob( "/Projects/AutomatorTest/raw/*.gz" )
#    for zipfile in unzipFiles:
#        print( "Unzipping: " + zipfile )
#        zorro.util.decompressFile( zipfile, outputDir="/Projects/AutomatorTest", n_threads=24 )
#    
#    # https://pymotw.com/2/multiprocessing/communication.html
#    skulky = skulkManager()
#    
#    skulky.zorroDefault.shapePadded = [2048,2048]
#    skulky.zorroDefault.fouCrop = [1024,1024]
#    skulky.zorroDefault.doCompression = True
#    
#    skulky.paths['input_dir'] = "/Projects/AutomatorTest" 
#    skulky.paths['sum_subdir'] = "/Projects/AutomatorTest/sum"
#    skulky.paths['align_subdir'] = "/Projects/AutomatorTest/align"
#    skulky.paths['raw_subdir'] = "/Projects/AutomatorTest/raw"
#    
#    
#    skulky.hostInfo['localhost1']= skulkHost( 'localhost1', 'local', 
#                skulky.hostMessageQueue, n_threads=12)
#    skulky.hostInfo['localhost2']= skulkHost( 'localhost2', 'local', 
#                skulky.hostMessageQueue, n_threads=12 )
#                
#    skulky.mainPollingLoop( sleepTime = 2.0 )
    pass
    
