# -*- coding: utf-8 -*-
"""
Fourier cropping (binning)

Created on Thu Apr  7 13:51:33 2016
@author: Robert A. McLeod

This is a low-overhead utility for cropping data in Fourier space, typically from 4k to 8k.  Counts are 
conserved (so the intensity per pixel will increase by binFactor **2 )

Call as:

python binDM4.py [filename] [binFactor] [n_threads] [fftw_wisdom]

Defaults are: binFactor = 2, n_threads = 16, fftw_wisdom = None

"""

from zorro import ioMRC, ioDM, util
import numpy as np
import pyfftw
import os, os.path
#import psutil, time

#filename = "8K_2016-03-23_03_55_39.dm4"
#n_threads = 16
#binFactor = 2
#wisdom_file = "/scratch/fftw_wisdom.pkl"

def binDM4( filename, binFactor = 2, n_threads = 16, wisdom_file = None ):
    #mem0 = psutil.virtual_memory()[2] * 128.0 / 100.0
    #t0 = time.time()
    dm4Struct = ioDM.DM4Import( filename )
    dm4Image = dm4Struct.im[1].imageData
    
    #t1 = time.time()
    #mem1 = psutil.virtual_memory()[2] * 128.0 / 100.0
    
    mageShape = np.array( [dm4Image.shape[1], dm4Image.shape[2]] )
    binShape = (mageShape / binFactor ).astype('int')
    
    binArray = np.zeros( [dm4Image.shape[0], binShape[0], binShape[1]], dtype='float32' )
    
    FFT2, _ = util.pyFFTWPlanner( dm4Image[0,:,:].astype('complex64'), 
                                 effort='FFTW_MEASURE', wisdomFile=wisdom_file, n_threads = n_threads, 
                                 doForward=True, doReverse=False  )
    _, IFFT2bin = util.pyFFTWPlanner( binArray[0,:,:].astype('complex64'), 
                                     effort='FFTW_MEASURE', wisdomFile=wisdom_file, n_threads = n_threads, 
                                     doForward=False, doReverse=True  )
    
    #t2 = time.time()
    #mem2 = psutil.virtual_memory()[2] * 128.0 / 100.0
    
    FFTImage = np.zeros( mageShape, dtype='complex64' )
    FFTBinned = np.zeros( binShape, dtype='complex64' )
    ImageBinned = np.zeros( binShape, dtype='complex64' )
    for J in np.arange( dm4Image.shape[0] ):
        # print( "%d, mean: %.2f" %(J, np.mean(dm4Image[J,:,:] ) ) )
        FFT2.update_arrays( dm4Image[J,:,:].astype('complex64'), FFTImage ); FFT2.execute()
        # Crop
        FFTBinned[:binShape[0]/2,:binShape[1]/2] = FFTImage[:binShape[0]/2,:binShape[1]/2]
        FFTBinned[:binShape[0]/2, -binShape[1]/2:] = FFTImage[:binShape[0]/2, -binShape[1]/2:]
        FFTBinned[-binShape[0]/2:,:binShape[1]/2] = FFTImage[-binShape[0]/2:,:binShape[1]/2]
        FFTBinned[-binShape[0]/2:,-binShape[1]/2:] = FFTImage[-binShape[0]/2:,-binShape[1]/2:]
        
        # Normalize
        FFTBinned /= binShape[0]*binShape[1]
        
        # Invert
        IFFT2bin.update_arrays( FFTBinned, ImageBinned ); IFFT2bin.execute()
        
        # De-complexify
        binArray[J,:,:] = np.real( ImageBinned )
        #print( "%d, binned mean: %.2f" %(J, np.mean(binArray[J,:,:] ) ) )
        pass
    
    #t3 = time.time()
    #mem3 = psutil.virtual_memory()[2] * 128.0 / 100.0
    
    fileFront = os.path.splitext( filename )[0]
    binName = fileFront + ".bin.mrc"
    # Convert DM4 pixelsize in nm to Angstroms for MRC format, and apply binFactor
    ioMRC.MRCExport( binArray, binName, pixelsize=dm4Struct.im[1].imageInfo['DimXScale']*10.0*binFactor )
    
    #t4 = time.time()
    #mem4 = psutil.virtual_memory()[2] * 128.0 / 100.0
    
    #print( "t : %.3f, mem : %.1f GB" % (t0-t0,mem0) )
    #print( "t : %.3f, mem : %.1f GB" % (t1-t0,mem1) )
    #print( "t : %.3f, mem : %.1f GB" % (t2-t1,mem2) )
    #print( "t : %.3f, mem : %.1f GB" % (t3-t2,mem3) )
    #print( "t : %.3f, mem : %.1f GB" % (t4-t3,mem4) )

if __name__ == "__main__":
    import sys

    try:
        filename = sys.argv[1]
    except:
        raise IOError( "No filename provided" )
    try: 
        binFactor = sys.argv[2]
    except:
        binFactor = 2
    try: 
        n_threads = sys.argv[3]
    except: 
        n_threads = 16
    try: 
        fftw_wisdom = sys.argv[4]
    except:
        fftw_wisdom = None
    
    if os.path.isfile( filename ):
        binDM4( filename, np.float32( binFactor ), np.int( n_threads ), fftw_wisdom )
    else:
        print( "File not found: %s, exiting" % filename )