# -*- coding: utf-8 -*-
"""
DM4Import.py

Created on Wed Apr 01 17:58:48 2015
@author: Robert A. McLeod
@email: robbmcleod@gmail.com OR robert.mcleod@unibas.ch

This is designed to be a fast DM4 file reader to strip the data out of the large
movie-mode files generated by K2 detectors along with important tags.  Due to 
the emphasis on speed it's not a general file parser with dynamic allocation and such, 
so if Gatan changes the format a lot it will break the script.

Chris Boothroyd provides some information on the structure of .DM4 files here:

http://www.er-c.org/cbb/info/dmformat/

The directory structure of the file is tracked by passing the path of the parent to 
each tag parser. Gatan likes to have empty tag directories, so I autonumber them.

Can examine file structure by setting verbose = True, if you need to add some more 
tags to the dictionary generated by the parseTag function, for example

"""

import numpy as np
from time import time
import zorro_util as util
# from plotting import ims


class DMImage:
    """
    From DM4Import.py
    
    This structure-like class is for storing 2D image or 3D image stack information, since 
    Digital Micrography likes to have more than one in a file.  Should store 
    Calibrations, Data, and ImageTags.  
    
    Data we will store in numpy format, for the other stuff we'll use a dictionary.
    
    Because of the general format we can also use this for MRC and other formats.
    """
    def __init__( self ):
        self.imageInfo = {} # Calibrations, tags, all in a dictionary
        self.shape = [] # necessary for unraveling
        self.imageData = np.zeros(1)
        self.imageSum = None
        self.FFTimage = None # Magnitude only
        self.polar = None
        self.FFTpolar = None # Magnitude only
        
        pass
    
    def sum( self, axis=0 ):
        print( "imageData.shape: " + str(self.imageData.shape) )
        if self.imageData.ndim == 2:
            self.imageSum = self.imageData
        else:    
            self.imageSum = np.sum( self.imageData, axis=axis )
        
    def FFT2( self, module = 'numpy' ):
        """
        Numpy Fourier transform
        """
        print( "DMImage computing FFT2" )
        if self.imageSum is None:
            self.sum()
        if module == 'numpy':
            self.FFTimage = np.abs( np.fft.fftshift( np.fft.fft2( self.imageSum ) ) )
        else:
            print( "DMImage: unknown module used for FFT" )
        
    def polarFFT2( self, module = 'numpy' ):
        """
        Polar Fourier transform using numpy (slowish)
        """
        print( "DMImage computing polarFFT2" )
        if self.FFTimage is None:
            self.FFT2( module = module )
        self.FFTpolar = util.img2polar( self.FFTimage )
        
class DM4Import:
    
    def __init__( self, filename, verbose=False, useMemmap = False ):
        # So basic strategy is use memmap to parse the tags and find the image data, which
        # can then be loaded fast with 
        f = open( filename, 'rb' )
        
        self.useMemmap = useMemmap
        
        # Instatiate class fields
        self._unnamedIndices = np.zeros( 20 ) # maximum number of nested tags
        self.im = [] # list of DMImage class objects
        self.verbose = verbose 
        
        # version = struct.unpack( '>I', f.read(4) )
        version = np.fromfile( f, dtype='>i4', count=1 )[0]
        if( version != 4):
            print( "Warning: Only DM4 is supported, will probably crash spectacularly" )
        np.fromfile( f, dtype='>i8', count=1 )[0] # rootlen
        byteord = np.fromfile( f, dtype='>i4', count=1 )[0] # 1 = little endian, generally it always is
        if byteord != 1:
            print( "Error: only little endian data ordering supported at present" )
            return
        
        np.fromfile( f, dtype='i1', count=1 )[0] # 1 = sorted
        np.fromfile( f, dtype='i1', count=1 )[0] # 1 = open structure
        ntags_root = np.fromfile( f, dtype='>i8', count=1 )[0] # number of tags
        
        # Now the tags start
        for J in np.arange( 0, ntags_root ):
            tag_type = np.fromfile( f, dtype='i1', count=1 )[0] # 20 = tag directory, 21 = tag base, 0 == EOF
        
            if( tag_type == 20 ):
                loc_tagstart = f.tell()
                tag_namelen = np.fromfile( f, dtype='>i2', count=1 )[0]
                tag_name = f.read( tag_namelen )
                tag_fieldlen = np.fromfile( f, dtype='>i8', count=1 )[0]  
                if tag_name == "ImageList":
                    # if self.verbose: print( "Found ImageList"  )
                    f.seek( loc_tagstart )
                    # So each image is an unnamed tag directory, composed of /ImageData and /ImageTags and /UniqueID
                    f = self.parseTagDir( f, "/" )
                    
                    # We currently do not care about any tag but ImageList
                    break
                    # Go to next tag in root directory
                    # f.seek( loc_ImageList + tag_fieldlen )
                    pass
                else:
                    # Go to next tag/tagdir
                    # print( "Seeking to: " + str(f.tell() + tag_fieldlen) )
                    f.seek( f.tell() + tag_fieldlen )
                
            elif( tag_type == 21 ):
                # throw away tags in root because we don't care for them
                f = self.discardTag( f )
            elif( tag_type == 0 ):
                break # EOF
            pass
        f.close()
        
        # Unravel images
        for mage in self.im:
            # FYI: Gatan stores image stacks [z,y,x] which is opposite to how they store the dimensions (but is Numpy and C convention)
            mage.shape = np.flipud( np.array( mage.shape ) )
            # This y-flip [:,::-1,...] puts us on the same origin-standard for data as MRC
            # Gatan uses a different standard for the origin location inside DM4 files.
            mage.imageData =  np.reshape( mage.imageData, mage.shape, order='C' )[:,::-1,...]
            # Image data is now compatible with MRCs saved in GMS, IMOD, EMAN2, etc.
        # Clean up anything else
            
    def parseTag( self, f, parent ):
        """ Parse a tag at the given file handle location """
        tag_namelen = np.fromfile( f, dtype='>i2', count=1 )[0]
        tag_name = f.read( tag_namelen )
        tag_fieldlen = np.fromfile( f, dtype='>i8', count=1 )[0]
        loc_tag = f.tell() # Save location so we can seek to end of tag later
        f.read(4) # Throw away %%%% seperator
        tag_ninfo = np.fromfile( f, dtype='>i8', count=1 )[0]
        if self.verbose: print( "Found tag: " + parent + tag_name + " with " + str(tag_ninfo) + " elements" )
            
        split_tag_name = (parent + tag_name).split('/')
        # Fixed nested parsing?
        # Skip ImageList
        imageIndex = int( split_tag_name[2] )
        # Check to see if self.im[imageIndex] exists
        try:
            self.im[imageIndex]
        except IndexError:
            self.im.append( DMImage() )    
            
        if split_tag_name[3] == 'ImageData':
            dimCount = 0
            if split_tag_name[4] == 'Calibrations':
                # No calibrations saved at present
                # /ImageList/1/ImageData/Calibrations/Brightness/Origin
                # /ImageList/1/ImageData/Calibrations/Brightness/Scale
                # /ImageList/1/ImageData/Calibrations/Brightness/Units
                # /ImageList/1/ImageData/Calibrations/Dimension/[0-2]/Origin
                # /ImageList/1/ImageData/Calibrations/Dimension/[0-2]/Scale
                # /ImageList/1/ImageData/Calibrations/Dimension/[0-2]/Units
                if split_tag_name[5] == 'Brightness':
                    if split_tag_name[6] == 'Origin':
                        self.im[imageIndex].imageInfo['IntensityOrigin'] = self.retrieveTagData( f, tag_ninfo )
                    elif split_tag_name[6] == 'Scale':
                        self.im[imageIndex].imageInfo['IntensityScale'] = self.retrieveTagData( f, tag_ninfo )  
                    elif split_tag_name[6] == 'Units':
#                        unicode_array = self.retrieveTagData( f, tag_ninfo )
#                        self.im[imageIndex].imageInfo['IntensityUnits'] = "".join([chr(item) for item in unicode_array])
                        self.im[imageIndex].imageInfo['IntensityUnits'] = self.retrieveTagData( f, tag_ninfo ).astype(np.uint8).tostring().decode('utf-8')
                        
                elif split_tag_name[5] == 'Dimension':
                    # 0 = x, 1 = y, 2 = z for split_tag_name[6]
                    if split_tag_name[6] == '0':
                        currDim = 'X'
                    elif split_tag_name[6] == '1':
                        currDim = 'Y'
                    elif split_tag_name[6] == '2':
                        currDim = 'Z'
                        
                    # Units is stored as uint16 array, which is actually unicode?
                    if split_tag_name[7] == 'Origin':
                        self.im[imageIndex].imageInfo['Dim'+currDim+'Origin'] = self.retrieveTagData( f, tag_ninfo )
                    elif split_tag_name[7] == 'Scale':
                        self.im[imageIndex].imageInfo['Dim'+currDim+'Scale'] = self.retrieveTagData( f, tag_ninfo )  
                    elif split_tag_name[7] == 'Units':
#                        unicode_array = self.retrieveTagData( f, tag_ninfo )
#                        self.im[imageIndex].imageInfo['Dim'+currDim+'Units'] = "".join([chr(item) for item in unicode_array])
                        # Sometimes we have wierd characters in the Units tag
                        try:
                            self.im[imageIndex].imageInfo['Dim'+currDim+'Units']= self.retrieveTagData( f, tag_ninfo ).astype(np.int16).tostring().decode('utf-8')
                        except:
                            try:
                                # "ISO-8859-1"
                                # 'windows-1252'
                                self.im[imageIndex].imageInfo['Dim'+currDim+'Units']= self.retrieveTagData( f, tag_ninfo ).astype(np.int16).tostring().decode("ISO-8859-1")
                            except:
                                pass
            
                pass
            elif split_tag_name[4] == 'Data':
                # self.verbose = True
                if self.verbose:
                    print( "Found image %d"%imageIndex + " at offset : " + str(f.tell() ) )
                self.im[imageIndex].imageData = self.retrieveTagData( f, tag_ninfo, memmap = self.useMemmap )
                # self.verbose = False
                pass
            elif split_tag_name[4] == 'Dimensions':
                self.im[imageIndex].shape.append( self.retrieveTagData( f, tag_ninfo ) )
                dimCount += 1
                pass
            #elif split_tag_name[4] == 'PixelDepth':
            #    pass
            pass
        elif split_tag_name[3] == 'ImageTags':
        
            if split_tag_name[4] == 'Acquisition':
                ### Find image flips ###
                # Here is where we can save any tags that we care for, like image rotations
                # /ImageList/1/ImageTags/Acquisition/Frame/Area/Transform/Transform List/0/Transpose/Horizontal Flip
                if split_tag_name[-1] == "Horizontal Flip" and split_tag_name[-4] == 'Device':
                    # So annoyingly there are several of these tags, only some of which are the true flips.
                    self.im[imageIndex].imageInfo['HorzFlip'] = self.retrieveTagData( f, tag_ninfo )
                elif split_tag_name[-1] == "Vertical Flip" and split_tag_name[-4] == 'Device':
                    self.im[imageIndex].imageInfo['VertFlip'] = self.retrieveTagData( f, tag_ninfo )
                elif split_tag_name[-1] == "Diagonal Flip" and split_tag_name[-4] == 'Device':   
                    self.im[imageIndex].imageInfo['DiagFlip'] = self.retrieveTagData( f, tag_ninfo )
                    
                # Pixel size seems to be complex, not sure if I'm retrieving it correctly (only important for non-square pixel detectors)
                elif split_tag_name[-1] == "Pixel Size (um)" and split_tag_name[-3] == 'Device':
                    self.im[imageIndex].imageInfo['DetectorPixelSize'] = self.retrieveTagData( f, tag_ninfo )
                elif split_tag_name[-1] == "Exposure" and split_tag_name[-2] == 'Detector':
                    self.im[imageIndex].imageInfo['Exposure'] = self.retrieveTagData( f, tag_ninfo )
                elif split_tag_name[-1] == "hbin" and split_tag_name[-2] == 'Detector':
                    self.im[imageIndex].imageInfo['HorzBin'] = self.retrieveTagData( f, tag_ninfo )
                elif split_tag_name[-1] == "vbin" and split_tag_name[-2] == 'Detector':
                    self.im[imageIndex].imageInfo['VertBin'] = self.retrieveTagData( f, tag_ninfo )
                
            elif split_tag_name[4] == 'Microscope Info':
                if split_tag_name[5] == 'Actual Magnification':
                    self.im[imageIndex].imageInfo['ActualMag'] = self.retrieveTagData( f, tag_ninfo )
                elif split_tag_name[5] == 'Indicated Magnification':
                    self.im[imageIndex].imageInfo['NominalMag'] = self.retrieveTagData( f, tag_ninfo )
                elif split_tag_name[5] == 'Cs(mm)':
                    self.im[imageIndex].imageInfo['C3'] = self.retrieveTagData( f, tag_ninfo )
                elif split_tag_name[5] == 'Voltage':
                    self.im[imageIndex].imageInfo['Voltage'] = self.retrieveTagData( f, tag_ninfo )   
                elif split_tag_name[5] == 'Mode':
                    self.im[imageIndex].imageInfo['MicroscopeMode'] = self.retrieveTagData( f, tag_ninfo ).astype(np.uint8).tostring().decode('ascii')  
                pass
            elif split_tag_name[4] == 'EELS Spectrometer':
                if split_tag_name[5] == 'Aperture index':
                    self.im[imageIndex].imageInfo['EELSApertureIndex'] = self.retrieveTagData( f, tag_ninfo ) 
                elif split_tag_name[5] == 'Aperture label':
                    self.im[imageIndex].imageInfo['EELSApertureLabel'] = self.retrieveTagData( f, tag_ninfo ).astype(np.uint8).tostring().decode('ascii')
                elif split_tag_name[5] == 'Energy loss (eV)':
                    self.im[imageIndex].imageInfo['EELSEnergyLoss'] = self.retrieveTagData( f, tag_ninfo )    
                elif split_tag_name[5] == 'Slit inserted':
                    self.im[imageIndex].imageInfo['EELSSlitIn'] = self.retrieveTagData( f, tag_ninfo )
                elif split_tag_name[5] == 'Slit width (eV)':
                    self.im[imageIndex].imageInfo['EELSSlitWidth'] = self.retrieveTagData( f, tag_ninfo )   
            elif split_tag_name[4] == "Latitude-S":
                if split_tag_name[5] == "Distance To Focus Position":
                    self.im[imageIndex].imageInfo['LatSDistanceToFocus'] = self.retrieveTagData( f, tag_ninfo )   
                elif split_tag_name[5] == "Intended Defocus":
                    self.im[imageIndex].imageInfo['LatSIntendedDefocus'] = self.retrieveTagData( f, tag_ninfo )      
                elif split_tag_name[5] == "Beam Diameter (nm)":
                    self.im[imageIndex].imageInfo['LatSBeamDiameter'] = self.retrieveTagData( f, tag_ninfo )         
        # Go to next location
        f.seek( loc_tag + tag_fieldlen )
        return f
        
    def getTagDType( self, bytecode ):
        if bytecode == 2: 
            return 'int16'
        elif bytecode == 3: 
            return 'int32'
        elif bytecode == 4: 
            return 'uint16'
        elif bytecode == 5:
            return 'uint32'
        elif bytecode == 6: 
            return 'float32'
        elif bytecode == 7: 
            return 'float64'
        elif bytecode == 8: 
            return 'bool'
        elif bytecode == 11:
            return 'int64'
        else:
            return ''
            
    def retrieveTagData( self, f, tag_ninfo, memmap = False ):
        """ Get the actual data from the tag, which is then stored in one of the dicts or imageData """
        # Here is basically where we should check tag_name to see if it's important...
        tag_loc = f.tell()
        tag_infos = np.fromfile( f, dtype='>i8', count=tag_ninfo )
        for K in np.arange(0,tag_infos.size):
            if tag_infos[K] == 15: # struct
                # Structs are organized really strangly in the file, they are essentially in the tag header so 
                # our reading doesn't work properly with them. In essence we need to go back to where 
                # we think the struct should be.  
                # print( "Correct struct ??? location: " + str( tag_loc + 8*K + 16 ) )
                f.seek( tag_loc + 8*K + 16 )
                struct_fieldcount = np.fromfile( f, dtype='>i8', count=1 )
                struct_fielddtype = []
                struct_fieldval = []
                for J in np.arange(0,struct_fieldcount):
                    np.fromfile( f, dtype='>i8', count=1 ) # Zero 
                    struct_fielddtype.append( self.getTagDType( np.fromfile( f, dtype='>i8', count=1 ) ) )
                    # print( str(J) + ": field dtype = " + str( struct_fielddtype[J] ) )
                for J in np.arange(0,struct_fieldcount):
                    struct_fieldval.append( np.fromfile(f, dtype=struct_fielddtype[J], count=1)[0] )
                    # print( str(J) + ": field val = " + str( struct_fieldval[J] ) )
                return struct_fieldval
            elif tag_infos[K] == 20: # array
                if self.verbose: print( "Found array" )
                try:
                    K += 1 # Warning: sometimes arrays are empty
                    array_dtype = self.getTagDType(tag_infos[K])
                    K += 1
                    array_ncount = tag_infos[K]
                    if self.verbose: print( "Array dtype: " + array_dtype + ", array ncount: " + str(array_ncount) )
                    if( array_ncount > 0 ):
                        if memmap:
                            return np.memmap( f, dtype=array_dtype, mode='r', shape=array_ncount )
                        else:
                            return np.fromfile( f, dtype=array_dtype, count=array_ncount )
                except IndexError:
                    pass
                pass
            elif tag_infos[K] == 9: # string
                tag_char = f.read(1)
                if self.verbose: print( "Found char: " + tag_char )
                pass
            elif tag_infos[K] == 18: # string
                # str_len = 
                if self.verbose: print( "FIXME Found string" )
                pass
            else: # singleton
                tag_dtype = self.getTagDType( tag_infos[K] )
                if tag_dtype != '':
                    tag_data = np.fromfile( f, dtype = tag_dtype, count=1 )[0]
                    # if self.verbose: print( "Singleton: " + tag_dtype + ": " + str(tag_data) ):
                    if self.verbose: print( "Singleton: " + tag_dtype + ": " + str(tag_data) )
                    return tag_data
        pass
    
    def discardTag( self, f ):
        """ Quickly parse to the end of tag that we don't care about its information """
        tag_namelen = np.fromfile( f, dtype='>i2', count=1 )[0]
        f.seek( f.tell() + tag_namelen )
        tag_fieldlen = np.fromfile( f, dtype='>i8', count=1 )[0]
        f.seek( f.tell() + tag_fieldlen )
        return f
        
    def parseTagDir( self, f, parent ):
        """ Parse a tag directory at the given file handle location """
        try:
            tag_namelen = np.fromfile( f, dtype='>i2', count=1 )[0]
            tag_name = f.read( tag_namelen )
            tag_fieldlen = np.fromfile( f, dtype='>i8', count=1 )[0]  
        except IndexError:
            if self.verbose: print( "Caught IndexError, trying to recover position in file" )
            # f.seek( f.tell() )
            return f
    
        # Handle empty tag_name by giving it an auto-generated number
        if tag_name == '':
            if( self.verbose ) : print( "Found empty tag" )
            tag_depth = int( parent.count( '/') - 1 )
            tag_name = '%d'%self._unnamedIndices[ tag_depth ]
            self._unnamedIndices[ tag_depth ] += 1
            # Reset all higher indices
            self._unnamedIndices[ tag_depth+1:] = 0
        loc_tagdir = f.tell()
        f.read( 2 ) # Throw away sorted and closed
        ntags_dir = np.fromfile( f, dtype='>i8', count=1 )[0]
        if self.verbose:
            print( "Found tag dir " + parent + tag_name + " with " + str(ntags_dir) + " tags" )
        
        # So typically ImageList has two empty name tag directories, of which one is various stuff and the other is the data
        for I in np.arange(0, ntags_dir ):
            try:
                subtag_type = np.fromfile( f, dtype='i1', count=1 )[0] # 20 = tag directory, 21 = tag base, 0 == EOF
            except IndexError:
                if self.verbose:
                    print( "Caught IndexError, trying to recover" )
                break
            if( subtag_type == 20 ):
                f = self.parseTagDir( f, parent + tag_name + "/" )
            elif( subtag_type == 21 ):
                f = self.parseTag( f, parent + tag_name + "/" )
            elif( subtag_type == 0 ):
                break # EOF
        # Go to next tag in root directory
        f.seek( loc_tagdir + tag_fieldlen )
        return f
    
    
        
        # End of DM4Import
        
#t0 = time()
#filename = "C:/temp/test1042015.dm4"
#test = DM4Import( filename, verbose=False )
#t1 = time()
#print( "Import time (s): " + str(t1-t0) )
#
## See
#print test.im[1].imageInfo
#from plotting import ims
#ims( test.im[1].imageData[0,:,:] )

